/* 

An abstract substitution is represented as (an equivalence class of) a
variable-monotyping (module equivalence relation induced by equality
relation on the domain of sets of substitutions). A variable-typing
maps a variable not in its domain to the type "top". The choice of a
partial function instead of a total function is non-essential. The
special symbol vtbot denotes the least abstract substitution.

*/

:- module(monotype, [ 
		     monotype_eq_constraints/2,
		     monotype_init/0,
		     monotype_list_print/1,
		     monotype_unify_entry/5,
		     monotype_unify_exit/6,
		monotype_bipai/3,
		monotype_compile/2,
		     monotype_bot/2,
		     monotype_ident/2,
		     monotype_join/3,
		     monotype_leq/2,
		     monotype_internalize/2,
		     monotype_externalize/2,
		monotype_bw_compile/2,
		monotype_bip_has_cond/2,
		     monotype_top/2,
		monotype_meet/3,
		monotype_bw_unify_entry/5,
		monotype_bw_unify_exit/6,
		monotype_bw_bipai/3
	       ]).



:- use_module(basics, [
	mgu/3,
	variable/1
   ]).

:- use_module(typedef, [typedef/3]).

:- use_module(library(lists)).

monotype_compile(Call,Call).

monotype_eq_constraints(Sub,EC) :- sort(Sub,EC).
monotype_internalize(ASub,ASub).
monotype_externalize(ASub,ASub).
monotype_list_print(_).
monotype_init.
monotype_unify_entry(Sub,Sigma,_NVarR,NVar,Zeta) :- 
           vt_tag(Sigma,NVar,SigmaR), 
	   monotype_solve(Sub,SigmaR,SigmaRSub),
           vt_restrict(SigmaRSub,NVar,Zeta).
monotype_unify_exit(Sub,Sigma,_NVarR,Theta,NVar,Zeta) :-
           vt_tag(Sigma,NVar,SigmaR), 
           monotype_comb(SigmaR,Theta,SigmaRTheta),
	   monotype_solve(Sub,SigmaRTheta,SigmaRThetaSub),
           vt_restrict(SigmaRThetaSub,NVar,Zeta).

monotype_bw_compile(_,_) :- 
    write("Backward analysis not supported"),nl,halt.
monotype_bip_has_cond(_,_) :- 
    write("Backward analysis not supported"),nl,halt.
monotype_meet(_,_,_) :- 
    write("Backward analysis not supported"),nl,halt.
monotype_bw_unify_entry(_,_,_,_,_) :- 
    write("Backward analysis not supported"),nl,halt.
monotype_bw_unify_exit(_,_,_,_,_,_) :- 
    write("Backward analysis not supported"),nl,halt.
monotype_bw_bipai(_,_,_) :- 
    write("Backward analysis not supported"),nl,halt.




vt_tag(vtbot,vtbot).
vt_tag([],_Shift,[]).
vt_tag(['$VAR'(Ni)/Ti|VTi],Shift,['$VAR'(No)/Ti|VTo]) :-
    No is Ni + Shift,
    vt_tag(VTi,Shift,VTo).


vt_restrict(vtbot,vtbot).
vt_restrict([],_NVar,[]).
vt_restrict(['$VAR'(Ni)/Ti|VTi],NVar,VTo) :-
    ( Ni =< NVar -> VTo = ['$VAR'(Ni)/Ti|VT] ; VTo=VT ), 
    vt_restrict(VTi,NVar,VT).




/* The combination of two abstract substitution for monotypemorphic type
   analysis.
*/

monotype_comb(vtbot,_,vtbot) :-
        !.
monotype_comb(_,vtbot,vtbot) :-
        !.
monotype_comb(VT1,VT2,VTo)	:-
        append(VT2,VT1,VTo). % VT1 constrains renamed variables

/* The identity abstract substitution is the empty variable-typing.
*/

monotype_ident(_NVars,[]).

/* The top abstract substitution is the empty variable-typing. */

monotype_top(_NVars,[]).

/* The bottom abstract substitution is the special symbol vtbot. */

monotype_bot(_NVars, vtbot).


/* monotype_ext(+VTi,+Vars,-VTo) VTo=VTi since default type assigned to a
   variable by a variable-typing is "top" */

monotype_ext(VT,_,VT).


/* monotype_rest(+VTi,+Vars,-VTo) VTo is the restriction of VTi to
   program variables
*/



monotype_rest(vtbot,vtbot).
monotype_rest([],[]).
monotype_rest([V/T|VTi],VTo)  :-
        vt_restrict([V/T|VTi],VTo).


/* monotype_solve(+Sub, +ASubi, -ASubo) -ASubo approximates the set of
   substitutions that satisfy both constraints +Sub and +ASubi. */

monotype_solve(EQS,ASubi,ASubo)  :-
	mdown(EQS,ASubi,ASub1),
        ( equal_to_vtbot(ASub1) -> ASub2=vtbot; ASub2=ASub1 ),
        mup(EQS,ASub2,ASub3),
        ( equal_to_vtbot(ASub3) -> ASubo=vtbot; ASubo=ASub3 ).

equal_to_vtbot(VT) :- 
	some_bot_type(VT).

some_bot_type([_/Tp|VT]) :-
	Tp = bot -> true ; some_bot_type(VT).


% mdown(+EQS,+VTi,-VTo)

mdown(_,vtbot,vtbot) :-
        !.
mdown(EQS,VTi,VTo)   :-
        mdown_1(EQS,VTi,VTo).

mdown_1([],VT,VT).
mdown_1([LHS/RHS|EQS],VTi,VTo) :-
        ( mapsto(LHS,VTi,Tp) -> true; Tp=top ),
        mlvt(Tp,RHS,VT1),
        mdown_1(EQS,VTi,VT2),
        vtglb(VT1,VT2,VTo),
        !.

% mup(+EQS,+VTi,-VTo)

mup(_,vtbot,vtbot) :-
        !.
mup(EQS,VTi,VTo)   :-
        mup_1(EQS,VTi,VTo).

mup_1([],VTi,VTi).
mup_1([EQ|EQS],VTi,VTo) :-
        lhs_rhs([EQ|EQS],LHSs,RHSs),
        map_mlt(RHSs,VTi,TpTup),
        lhs_rhs(VTo1,LHSs,TpTup),
        vtglb(VTi,VTo1,VTo),
        !.

        lhs_rhs([],[],[]).
        lhs_rhs([LHS/RHS|EQS],[LHS|LHSs],[RHS|RHSs]) :-
                  lhs_rhs(EQS,LHSs,RHSs).

/* monotype_bipai(+Call,+ASubi,-ASubo) most built-in predicates do not
   change program state, that is, ASubo=ASubi if +Call succeeds.
   Much of monotype_bipai/3 can be improved. */

monotype_bipai(repeat,          ASubi,          ASubi).
monotype_bipai(listing,         ASubi,          ASubi).
monotype_bipai(listing(_Term),  ASubi,          ASubi).
monotype_bipai(compound(_Term), ASubi,          ASubi).
monotype_bipai(portray_clause(_),ASubi,          ASubi).
monotype_bipai(float(_Term),    ASubi,          ASubi).
monotype_bipai(string(_Term),   ASubi,          ASubi).
monotype_bipai(ground(_Term),   ASubi,          ASubi).
monotype_bipai(compare(_,_,_),  ASubi,          ASubi).

monotype_bipai(!,	        ASubi,		ASubi).
monotype_bipai(display(_Term),	ASubi,		ASubi).
monotype_bipai(put(_Term),	ASubi,		ASubi).
monotype_bipai(nl,            	ASubi,		ASubi).
monotype_bipai(print(_Term),	ASubi,        	ASubi).
monotype_bipai(tab(_Natural),	ASubi,      	ASubi).
monotype_bipai(write(_Term),	ASubi,        	ASubi).
monotype_bipai(writeq(_Term),	ASubi,        	ASubi).
monotype_bipai(=..(_,_),	ASubi,        	ASubi).
monotype_bipai(functor(_,_,_),	ASubi,        	ASubi).
monotype_bipai(length(_,_),	ASubi,        	ASubi).
monotype_bipai(arg(_,_,_),	ASubi,        	ASubi).

monotype_bipai(atom(_Term),		ASubi,        	ASubi).
monotype_bipai(atomic(_Term),	ASubi,        	ASubi).
monotype_bipai(integer(_Term),	ASubi,        	ASubi).
monotype_bipai(number(_Term),	ASubi,        	ASubi).

monotype_bipai(statistics(_N1,_N2),	ASubi,        	ASubi).
monotype_bipai(name(_N1,_N2),	ASubi,        	ASubi).
monotype_bipai(sort(_N1,_N2),	ASubi,        	ASubi).
monotype_bipai(keysort(_N1,_N2),	ASubi,        	ASubi).
monotype_bipai(=:=(_N1,_N2),	ASubi,        	ASubi).
monotype_bipai(=\=(_N1,_N2),	ASubi,        	ASubi).
monotype_bipai(<(_N1,_N2),	ASubi,        	ASubi).
monotype_bipai(>(_N1,_N2),	ASubi,        	ASubi).
monotype_bipai(=<(_N1,_N2),	ASubi,        	ASubi).
monotype_bipai(>=(_N1,_N2),	ASubi,        	ASubi).
monotype_bipai(@<(_T1,_T2),	ASubi,        	ASubi).
monotype_bipai(@>(_T1,_T2),	ASubi,        	ASubi).
monotype_bipai(@=<(_T1,_T2),	ASubi,        	ASubi).
monotype_bipai(@>=(_T1,_T2),	ASubi,        	ASubi).

monotype_bipai(true,		ASubi,        	ASubi).
monotype_bipai(nonvar(_Term),	ASubi,          ASubi).

monotype_bipai(abort,		_ASubi,		ASubo) 	:-
    	monotype_bot([],ASubo).
monotype_bipai(fail,		_ASubi,		ASubo) 	:-
    	monotype_bot([],ASubo).
monotype_bipai(false,		_ASubi,		ASubo) 	:-
    	monotype_bot([],ASubo).
monotype_bipai(var(_Variable),	ASubi,		ASubi).
monotype_bipai(is(_Left,_Right),	ASubi,		ASubi).
monotype_bipai(read(_Term),		ASubi,    	ASubi).

monotype_bipai(=(Left,Right),	ASubi,		ASubo) 	:-
	mgu(Left,Right,E0), 
	monotype_solve(E0,ASubi,ASubo).
monotype_bipai(=@=(Left,Right),     ASubi,          ASubo)  :-
	monotype_bipai(=(Left,Right), ASubi, ASubo). 
	% identical terms must unify
monotype_bipai(==(Left,Right),     ASubi,          ASubo)  :-
	monotype_bipai(=(Left,Right), ASubi, ASubo). 
	% identical terms must unify
monotype_bipai(\==(_Left,_Right),     ASubi,          ASubi).
monotype_bipai(\=(_,_),       ASubi,          ASubi).



/* tleq(+T1,+T2) partial order on monotype-types*/

tleq(bot,_) :-             % "bot" is the minimum type
    	!.
tleq(_,top) :-             % "top" is the top type
    	!.
tleq(T1,T2) :-
%   	 \+ T1==bot,
%   	 \+ T2==top,
    	T1 =.. [Cons|Types1],
    	T2 =.. [Cons|Types2],
    	map_tleq(Types1,Types2).

map_tleq([],[]).
map_tleq([T1|L1],[T2|L2]) :-
    tleq(T1,T2),
    map_tleq(L1,L2).

tlub(top,_,top) :-!.
tlub(_,top,top) :-!.
tlub(bot,T,T) :-!.
tlub(T,bot,T) :-!.
tlub(T1,T2,T3)  :-
	!,
    	(
        	functor(T1,Cons,Arity), 
		functor(T2,Cons,Arity) 
    	->
        	T1 =.. [Cons|PARs1], 
		T2 =.. [Cons|PARs2],
        	map_tlub(PARs1,PARs2,PARs3),
        	T3 =.. [Cons|PARs3]
    	;
		T3 = top
    	).

map_tlub([],[],[]).
map_tlub([T1|PARs1],[T2|PARs2],[T3|PARs3]) :-
    	tlub(T1,T2,T3),
    	map_tlub(PARs1,PARs2,PARs3).


/*tglb(+T1,+T2,-T3) -T3 is the intersection of type +T1 and +T2.*/

tglb(bot,_,bot)	:-!.
tglb(_,bot,bot) :-!.
tglb(top,T,T)	:-!.
tglb(T,top,T)	:-!.
tglb(T1,T2,T3) 	:-
    	(
        	functor(T1,Cons,Arity),
	        functor(T2,Cons,Arity) 
	->
        	T1 =.. [Cons|Args1],
	        T2 =.. [Cons|Args2],
        	map_tglb(Args1,Args2,Args3),
	        T3 =.. [Cons|Args3]
	;
        	T3 = bot
	).

map_tglb([],[],[]).
map_tglb([T1|Args1],[T2|Args2],[T3|Args3]) 	:-
	tglb(T1,T2,T3),
    	!,
    	map_tglb(Args1,Args2,Args3).

/* vtglb(+VT1,+VT2,-VT3) +VT1 and +VT2 are variable-typings. If one of
   them is vtbot then -VT3 is vtbot. Otherwise vtglb/3 succeeds with
   dom(VT3)=dom(VT1)+dom(VT2) and for each varaible X in dom(VT3) such
   that X/T is in VT1 and X/T2 in VT2, X/T3 is in VT3 and T3 is the
   greatest lower bound of T1 and T2.

   This implementation assumes that variable typings are sorted lists
   that are sorted according to the ordering on the names of
   variables.*/

vtglb(vtbot,_,vtbot)    :-
        !.
vtglb(_,vtbot,vtbot)    :-
        !.
vtglb([],VT,VT) 	:-
        !.
vtglb(VT,[],VT) 	:-
        !.
vtglb([X1/T1|VT1],[X2/T2|VT2],[X1/T3|VT3]) 	:-
    	X1 == X2,
    	!,
    	tglb(T1,T2,T3),
    	vtglb(VT1,VT2,VT3).
vtglb([X1/T1|VT1],[X2/T2|VT2],[X1/T1|VT3]) 	:-
    	X1 @< X2,
    	!,
    	vtglb(VT1,[X2/T2|VT2],VT3).
vtglb([X1/T1|VT1],[X2/T2|VT2],[X2/T2|VT3]) 	:-
    	X1 @> X2,
    	!,
    	vtglb([X1/T1|VT1],VT2,VT3).


/* mlvt(+Tp,+Tm,-VT) always succeeds with maximum VT such as
   Tp>=VT(Tm) */

mlvt(top,_,[])          :-
    	!.
mlvt(bot,_,vtbot)	:-
	!.
mlvt(Tp,Var,[Var/Tp])	:-
%	\+ Tp = top, \+ Tp = bot,
	variable(Var),
	!.
mlvt(Tp,Tm,VTo) 	:-
%	\+ Tp = top, \+ Tp = bot, 
%       \+ variable(Tm),
    	functor(Tm,Func,Arity),
	typedef(Func/Arity,Tis,To),
	(
		Tp = To
	->
    		Tm  =.. [_|Tms],
		map_mlvt(Tis,Tms,VTo)
	;
		VTo = vtbot
	),
	!.

/* map_mlvt/3 - distributed version of mlvt/3 */

map_mlvt([],[],[]).
map_mlvt([Tp|Tps],[Tm|Tms],VTo) :-
    	mlvt(Tp,Tm,VT1),
    	map_mlvt(Tps,Tms,VT2),
	vtglb(VT1,VT2,VTo).

/* mtsublub(+TSub1,+TSub2,-TSubo) */

mtsublub(fail,	_,	fail)	:-
	!.
mtsublub(_,	fail,	fail)	:-
	!.
mtsublub(TSub1,	TSub2,	TSubo)	:-
	mtsublub_1(TSub1,TSub2,TSubo).

mtsublub_1([],[],[]).
mtsublub_1([],[P/T|TSub],[P/T|TSub]).
mtsublub_1([P/T|TSub],[],[P/T|TSub]).
mtsublub_1([P1/T1|TSub1],[P2/T2|TSub2],[P/T|TSub]) :-
	(
		P1 @< P2
	->
		P=P1,
		T=T1,
		mtsublub_1(TSub1,[P2/T2|TSub2],TSub)
	;
		P1 @> P2
	->
		P=P2,
		T=T2,
		mtsublub_1([P1/T1|TSub1],TSub2,TSub)
	;
		P=P1, 
		tlub(T1,T2,T),
		mtsublub_1(TSub1,TSub2,TSub)
	).

/* tsub_cons(+Paras,+Tps,-TSub) */

tsub_cons(Paras,Tps,TSub) :- 
	tsub_cons_1(Paras,Tps,TSubM),
    	sort(TSubM,TSub).

tsub_cons_1([],[],[]).
tsub_cons_1([Para|Paras],[Tp|Tps],[Para/Tp|TSub]) :-
    	tsub_cons_1(Paras,Tps,TSub).

/* parameter(+Tp) iff Tp is a parameter in type rules */

parameter(Tp) :- var(Tp).


/* 

mapsto(+Key,+Map,-Value) If +Key is in the domain of +Map
then mapsto(Key,Map,Value) succeeds and Value is the value
assigned to Key by Map. Otherwise mapsto(Key,Map,Value)
fails.

*/

mapsto(Key,[Key1/Value1|Map],Value) :-
	( Key==Key1 -> Value=Value1 ; mapsto(Key,Map,Value) ).

/*

tsub_apply(+TSub,+TermIn,-TermO) TermO=TSub(TermIn). It
unravals the input term TermIn in the top-down manner and
constructs the output term TermO in the bottom-up manner.

*/
tsub_apply(fail,_,top) :- !.

tsub_apply(TSub,TpIn,TpO) :-
	(    
		var(TpIn) %parameter(TpIn)
	->
		( 
			mapsto(TpIn,TSub,TpO) 
		-> 
			true
		; 
			TpO=bot % note default value for any
				% Tp parameter is bot
		)
	; 
		TpIn =.. [F|ArgsIn],
		map_tsub_apply(ArgsIn,TSub,ArgsO),
		TpO =.. [F|ArgsO]         
	).

map_tsub_apply([],_,[]).
map_tsub_apply([ArgIn|ArgsIn],TSub,[ArgO|ArgsO]) :-
	tsub_apply(TSub,ArgIn,ArgO),
	map_tsub_apply(ArgsIn,TSub,ArgsO).


/* map_mlt/3 - distributed version of plt/3 */

map_mlt([],_,[]).
map_mlt([Tm|Tms],VT,[Tp|Tps]) :-
    	mlt(Tm,VT,Tp),
    	map_mlt(Tms,VT,Tps).

/* mlt(+Tm,+VT,-Tp) succeeds with -Tp>=VT(Tm). */

mlt(V,VT,Tp) :-
    	variable(V),
    	!,
	( mapsto(V,VT,Tp) -> true ; Tp = top ).

mlt(F0,_,Tp)  :-
	atomic(F0),
	!,
	typedef(F0/0,[],GTp),
	tsub_apply([],GTp,Tp).

mlt(Tm,VT,Tp) :-
%	\+ variable(V),
%	\+ atomic(F0),
    	functor(Tm,Func,Arity),
	typedef(Func/Arity,Tins,To),
    	Tm  =.. [_|Tms],
    	map_mlt(Tms,VT,Tps),
	map_mlts(Tps,Tins,TSub),
	tsub_apply(TSub,To,Tp),
	!.

/* mlts(+Tp,+GTp,-TSub) succeeds with -TSub being the minimum solution
   for Tp=<TSub(GTp).  */

mlts(_,   GTp,  [])	:-
        GTp==top,  % substituting GTp with "top" is a BUG! 
    	!.
mlts(Tp,GTp,[GTp/Tp])	:-  
%	GTp \== top,
    	var(GTp), %parameter(GTp),
    	!.

mlts(Tp,GTp,TSub) 	:-
%    	GTp \== top,
%	\+ var(GTp), %\+ parameter(GTp),
    	(
		(functor(Tp,Cons,Arity),functor(GTp,Cons,Arity))
	->
		Tp =.. [_|Tps], 
		GTp =.. [_|Paras],
		tsub_cons(Paras,Tps,TSub)
	;
		TSub = fail
	),
	!.

map_mlts([],[],[]).
map_mlts([Tp|Tps],[GTp|GTps],TSub) 	:-
    	mlts(Tp,GTp,TSub1),
    	map_mlts(Tps,GTps,TSub2),
    	mtsublub(TSub1,TSub2,TSub),
	!.


/* monotype_leq(+VT1,+VT2) VT1 is less than VT2. */

monotype_leq(ASub1,ASub2)  :-
        ( equal_to_vtbot(ASub1) -> VT1=vtbot; VT1=ASub1 ),
        ( equal_to_vtbot(ASub2) -> VT2=vtbot; VT2=ASub2 ),
        vtleq(VT1,VT2).


/* vtleq(+VT1,+VT2) +VT1 and +VT2 are variable-typings. It is true
   when VT1 is vtbot. If VT2 is vtbot then VT1 must be vtbot.
   Otherwise vtleq/2 succeeds if  each varaible X in dom(VT1)
   VT1(X) is less than VT2(X)

   This implementation assumes that variable typings are sorted lists
   that are sorted according to the ordering on the names of
   variables.*/

vtleq(vtbot,_)      :-
        !.
vtleq(VT,vtbot)     :-
        !,
        VT == vtbot.
vtleq(_,[])           :-
        !.
vtleq([],[_V/Tp|VT])  :-
        !,
        Tp == top,
        vtleq([],VT).
        
vtleq([X1/T1|VT1],[X2/T2|VT2])  :-
    	X1 == X2, 
   	!,
    	tleq(T1,T2),
    	vtleq(VT1,VT2).
vtleq([X1/_T1|VT1],[X2/T2|VT2]) :-
    	X1 @< X2,
    	!,
    	vtleq(VT1,[X2/T2|VT2]).
vtleq([X1/T1|VT1],[X2/T2|VT2])  :-
    	X1 @> X2,
    	!,
        T2 == top,
    	vtleq([X1/T1|VT1],VT2).




% monotype_join(+CVTS1,+CVTS2,-CVTS) CVTS=CVTS1+CVTS2 redundency removed.

monotype_join(VT1, VT2, VTo) :-
        vtlub(VT1, VT2, VTo).

/* vtlub(+VT1,+VT2,-VT3) +VT1 and +VT2 are variable-typings. If one of
   them is vtbot then -VT3 is the other. Otherwise vtlub/3 succeeds
   with dom(VT3)=dom(VT1)/\dom(VT2) and for each varaible X in dom(VT3)
   such that X/T is in VT1 and X/T2 in VT2, X/T3 is in VT3 and T3 is
   the lub of T1 and T2.

   This implementation assumes that variable typings are sorted lists
   that are sorted according to the ordering on the names of
   variables.*/

vtlub(vtbot,VT,VT)      :-
        !.
vtlub(VT,vtbot,VT)      :-
        !.
vtlub([],_,[]) 	        :-
        !.
vtlub(_,[],[]) 	        :-
        !.
vtlub([X1/T1|VT1],[X2/T2|VT2],[X1/T3|VT3])  :-
    	X1 == X2, 
   	!,
    	tlub(T1,T2,T3),
    	vtlub(VT1,VT2,VT3).
vtlub([X1/_T1|VT1],[X2/T2|VT2],VT3) 	    :-
    	X1 @< X2,
    	!,
    	vtlub(VT1,[X2/T2|VT2],VT3).
vtlub([X1/T1|VT1],[X2/_T2|VT2],VT3) 	    :-
    	X1 @> X2,
    	!,
    	vtlub([X1/T1|VT1],VT2,VT3).


