/* 

A abstract substitution over domain V is a symmetric relation on V.
We require that all possible sharing in the input abstract
substitution be specified.  Note that variables in a clause are
numbered from 1 to NVar where NVar is the number of variables and NVar
is given in lieu of V.

*/


:- module(pair, [ 
		  pair_eq_constraints/2,
		  pair_init/0,
		  pair_list_print/1,
		  pair_unify_entry/5,
		  pair_unify_exit/6,
		  pair_bipai/3,
		  pair_compile/2,
		  pair_bot/2,
		  pair_ident/2,
		  pair_join/3,
		  pair_leq/2,
		  pair_internalize/2,
		  pair_externalize/2,
		  pair_bw_compile/2,
		  pair_bip_has_cond/2,
		  pair_top/2,
		  pair_meet/3,
		  pair_bw_unify_entry/5,
		  pair_bw_unify_exit/6,
		  pair_bw_bipai/3,
		  % following predicates are for experiments
	pair_solve_abduce/3,
        pair_ext/3,
	compose_sh/3,
	compose_sh_new/3
   ]).


:- use_module(library(lists)).

:- use_module(ordsets).

:- use_module(basics,	[
        variable/1,
        variable_id/2,
        vars/2,
        var_occ/2,
        mgu/3
   ]).


pair_bw_compile(_,_) :- 
    write("Backward analysis not supported"),nl,halt.
pair_bip_has_cond(_,_) :- 
    write("Backward analysis not supported"),nl,halt.
pair_bw_unify_entry(_,_,_,_,_) :- 
    write("Backward analysis not supported"),nl,halt.
pair_bw_unify_exit(_,_,_,_,_,_) :- 
    write("Backward analysis not supported"),nl,halt.
pair_bw_bipai(_,_,_) :- 
    write("Backward analysis not supported"),nl,halt.




pair_unify_entry(Sub,Sigma,_NVarR,NVar,Zeta) :- 
           pair_tagging(Sigma,NVar,SigmaR), 
	   pair_solve(Sub,SigmaR,SigmaRSub),
           pair_rest(SigmaRSub,NVar,Zeta).
pair_unify_exit(Sub,Sigma,_NVarR,Theta,NVar,Zeta) :-
           pair_tagging(Sigma,NVar,SigmaR), 
           pair_comb(SigmaR,Theta,SigmaRTheta),
	   pair_solve(Sub,SigmaRTheta,SigmaRThetaSub),
           pair_rest(SigmaRThetaSub,NVar,Zeta).




% no initialization to be done.

pair_init.

% only needed for POS

pair_list_print(_).


/* The identity abstract substitution on V is empty since
   there is no sharing between variables in V when everyone
   in V is instantiated into a distinct variable. */

pair_ident(_V,[]).


/* The top abstract substitution on V is V x V since there might be
   sharing between each pair of variables in V.  Note that variables
   in a clause are numbered from 1 to NVar where NVar is the number of
   variables and NVar is given in lieu of V. */

pair_top(NVar,S) :-
    integers(NVar,Vids),
    ord_setproduct(Vids,Vids,S).

    integers(N,L) :- integers(1,N,L).

    integers(I,N,[])    :- I>N.
    integers(I,N,[I|L]) :- I=<N, I1 is I+1, integers(I1,N,L).

/* The bottom abstract substitution on V is a special symbol
   "bot" for any V. */

pair_bot(_NVar, bot).

/* pair_leq(ASub1,ASub2) <==> ASub1=<ASub2 */

pair_leq(bot, _)   :-
	!.
pair_leq(ASub,bot) :-
	!,
	ASub==bot.
pair_leq(Sx,Sy)    :-
	ord_subset(Sx,Sy).

/* pair_meet(ASub1, ASub2, ASub3) <==> (ASub1/\ASub2)=ASub3
*/

pair_meet(bot,_ASub,bot) :-
	!.
pair_meet(_ASub,bot,bot) :-
	!.
pair_meet(Sx,Sy,Sxy)     :-
	ord_intersection(Sx,Sy,Sxy).


/* pair_join(ASub1, ASub2, ASub3) <==> (ASub1\/ASub2)=ASub3
*/

pair_join(bot,ASub,ASub) :-
	!.
pair_join(ASub,bot,ASub) :-
	!.
pair_join(Sx,Sy,Sxy)     :-
	ord_union(Sx,Sy,Sxy).

/* It is assumed that variables whose are not specified are
   free and different other free variables and from each
   other. */

pair_ext(ASub,_,ASub).

/* 

pair_ext(ASubi,V,ASubo) The sharing part is extended in
the following way. The extended sharing relation includes the
unextended sharing relation. The variables in V outside the
domain of ASubi may share with any non-ground variable in V.

*/

/*
pair_ext(bot,_,bot) :-
    !.
pair_ext(Si, V, So) :-
    sh_domain(Si, Vold),
	ord_subtract(V, Vold, Vnew),
	ord_setproduct(Vold,Vnew,Soldnew),
    ord_setproduct(Vnew,Vnew,Snewnew),
	ord_union(Si,Soldnew,S1),
    ord_union(S1,Snewnew,So).
*/

/* PRIVATE the domain of a sharing relation is the set of variables in
   it. The following implementation makes use of the fact that a
   sharing relation is represented as an order set 

*/

sh_domain([],[]).
sh_domain([(X,_)],[X]).
sh_domain([(X,_),(X,Y)|S],Vs) :-
    !,
    sh_domain([(X,Y)|S],Vs).
sh_domain([(X1,_),(X2,Y2)|S],[X1|Vs]) :- % X1\=X2
    sh_domain([(X2,Y2)|S],Vs).

/* PRIVATE derived_id_sh(+Si,-So). The derived identity of a sharing
   relation is the identify relation over the domain of the sharing
   relation. The following implementation makes use of the fact that a
   sharing relation is represented as an order set. */

derived_id_sh([],[]).
derived_id_sh([(X,_)],[(X,X)]).
derived_id_sh([(X,_),(X,Y)|Si],So) :-
    !,
    derived_id_sh([(X,Y)|Si],So).
derived_id_sh([(X1,_),(X2,Y2)|Si],[(X1,X1)|So]) :- % X1\=X2
    derived_id_sh([(X2,Y2)|Si],So).


/* pair_rest(ASubI,NVar,ASubO) <==> ASubO is the restriction of
   ASubI to program variables numbered 1 to NVar. 
*/

pair_rest(bot,_,bot).
pair_rest([],_,[]).
pair_rest([(X,Y)|Si],NVar,[(X,Y)|So]) :-
        X =< NVar,
        Y =< NVar,
        !,
        pair_rest(Si,NVar,So).

pair_rest([_XY|Si],NVar,So) :- 
%       \+( X =< NVar, Y =< NVar ),
        pair_rest(Si,NVar,So).



/* 
PRIVATE unieq_sh(+Kx,+Kt,+X,+Vt,-So) Kx=\multiplicity(X,ASub)  and
Kt=\multiplicity(t,ASub) are pre-computed. 



unieq_sh(0,_,_,_,[]).
unieq_sh(_,0,_,_,[]).
unieq_sh(1,1,X,Vt,So) :-
    ord_setproduct([X],Vt,XVt),
    ord_setproduct(Vt,[X],VtX),
    ord_union(XVt,VtX,So).
unieq_sh(1,2,X,Vt,So) :-
    ord_setproduct([X],Vt,XVt),
    ord_setproduct(Vt,[X],VtX),
    ord_union(XVt,VtX,So1),
    ord_add_element(So1,(X,X),So).
unieq_sh(2,1,X,Vt,So) :-
    ord_setproduct([X],Vt,XVt),
    ord_setproduct(Vt,[X],VtX),
    ord_union(XVt,VtX,So1),
    ord_setproduct(Vt,Vt,Vt2),
    ord_union(So1,Vt2,So).
unieq_sh(2,2,X,Vt,So) :-
    ord_setproduct([X],Vt,XVt),
    ord_setproduct(Vt,[X],VtX),
    ord_union(XVt,VtX,So1),
    ord_setproduct(Vt,Vt,Vt2),
    ord_union(So1,Vt2,So2),
    ord_add_element(So2,(X,X),So).
*/




unieq_sh(0,_,_,_,[]).
unieq_sh(_,0,_,_,[]).
unieq_sh(1,1,X,Vt,So) :-
    left_right(X,Vt,So).
unieq_sh(1,2,X,Vt,So) :-
    left_right_xx(X,Vt,So).
unieq_sh(2,1,X,Vt,So) :-
    left_right(X,Vt,So1),
    ord_setproduct(Vt,Vt,Vt2),
    ord_union(So1,Vt2,So).
unieq_sh(2,2,X,Vt,So) :-
    left_right_xx(X,Vt,So1),
    ord_setproduct(Vt,Vt,Vt2),
    ord_union(So1,Vt2,So).


/* PRIVATE left_right(+X,+Vs,-S) iff S={X}Vs U Vs{X} We divide Vs into
   two subsequences -first one containing Ys that are smaller than X
   and second one containing Zs that are greater than X. S will have
    three segments - Ys{X}#{X}Zs#Zs{X}
*/

left_right(X,Vs,Hf) :- left_right_1(Vs,X,Hf,Hm,Hm). 

left_right_1([],_,S,S,[]). 
left_right_1([Y|Vs],X,Tf,Hm,Tm) :- 
    Y < X,
    !,
    Tf=[(Y,X)|TfNew],
    Tm=[(X,Y)|TmNew],
    left_right_1(Vs,X,TfNew,Hm,TmNew). 

left_right_1([Y|Vs],X,Hm,Hm,Tm) :- 
%    Y >= X,
    left_right_2([Y|Vs],X,Tm,Hr,Hr). 

left_right_2([],_,Tm,Tm,[]). 
left_right_2([Z|Vs],X,Tm,Hr,Tr) :-
    Tm=[(X,Z)|TmNew],
    Tr=[(Z,X)|TrNew],
    left_right_2(Vs,X,TmNew,Hr,TrNew).



/* PRIVATE left_right_xx(+X,+Vs,-S) iff S={X}Vs U Vs{X} U{(X,X)} We
   divide Vs into two subsequences -first one containing Ys that are
   smaller than X and second one containing Zs that are greater than
   X. S will have three segments - Ys{X}#{(X,X)}#{X}Zs#Zs{X} */

left_right_xx(X,Vs,Hf) :- left_right_xx_1(Vs,X,Hf,Hm,Hm). 

left_right_xx_1([],X,S,S,[(X,X)]). 
left_right_xx_1([Y|Vs],X,Tf,Hm,Tm) :- 
    Y < X,
    !,
    Tf=[(Y,X)|TfNew],
    Tm=[(X,Y)|TmNew],
    left_right_xx_1(Vs,X,TfNew,Hm,TmNew). 

left_right_xx_1([Y|Vs],X,Hm,Hm,[(X,X)|Tm]) :- 
%    Y >= X,
    left_right_2([Y|Vs],X,Tm,Hr,Hr). 



/* 

PRIVATE compose_sh(+Sin,+Seq,-So) -So is the  composition of
sharing  relations  Sin  and  Seq wher  Seq is  the  sharing
relation on the varaibles in the equation X=T.

*/

compose_sh(Sin,Seq,So) :-
    relation_join(Seq,Sin,SeqSin),
    ord_union(Seq,SeqSin,T),
    relation_join(Sin,T,SinT),
    ord_union(Sin,SinT,U),
    ord_union(T,U,So).


/* PRIVATE/OPTIMIZED. relation_join(+R1,+R2,-R3) (x,z) is in R3 iff
   there is y such that (x,y) is in R1 and (y,z) is in R2. The
   implementation assumes that a relation is represented as an ordered
   list. 

relation_join(R1,R2,Ro) :-
    relation_join(R1,R2,R2,[],Ro).

relation_join([],_,_,Ro,Ro).
relation_join([(X,Y)],R2,_,Racc,Ro) :-
%    !,
    relation_join_1(R2,X,Y,R2XY,_),
    ord_union(R2XY,Racc,Ro).
relation_join([(X,Y),(X,Y1)|R1],R2,R2F,Racc,Ro) :-
    !,
    relation_join_1(R2,X,Y,R2XY,R2New),
    ord_union(R2XY,Racc,Racc1),
    relation_join([(X,Y1)|R1],R2New,R2F,Racc1,Ro).
relation_join([(X,Y),(X1,Y1)|R1],R2,R2F,Racc,Ro) :- %X \= Y
    relation_join_1(R2,X,Y,R2XY,_),
    ord_union(R2XY,Racc,Racc1),
    !,
    relation_join([(X1,Y1)|R1],R2F,R2F,Racc1,Ro).

*/

relation_join(R1,R2,Ro) :-
    relation_join(R1,R2,R2,[],Ro,[]). % using difference list

relation_join([],_,_,_,Ro,Ro).

relation_join([(X,Y)],R2,_,RxAcc,Front,Tail) :-
%    !,
    relation_join_1(R2,X,Y,R2XY,_),
    ord_union(R2XY,RxAcc,Rx),
    append(Rx,Tail,Front).

relation_join([(X,Y),(X,Y1)|R1],R2,R2F,RxAcc,Front,Tail) :-
    !,
    relation_join_1(R2,X,Y,R2XY,R2New),
    ord_union(R2XY,RxAcc,RxAccNew),
    relation_join([(X,Y1)|R1],R2New,R2F,RxAccNew,Front,Tail).

relation_join([(X,Y),(X1,Y1)|R1],R2,R2F,RxAcc,Front,Tail) :- %X \= Y
    relation_join_1(R2,X,Y,R2XY,_),
    ord_union(R2XY,RxAcc,Rx),
    append(Rx,Front1,Front),
    !,
    relation_join([(X1,Y1)|R1],R2F,R2F,[],Front1,Tail).


relation_join_1([],_,_,[],[]).
relation_join_1([(Y,Z)|Ri],X,Y,[(X,Z)|Ro],Remainder) :-
    !,    
    relation_join_1(Ri,X,Y,Ro,Remainder).
relation_join_1([(Y1,_)|Ri],X,Y,Ro,Remainder) :-
    Y1 @< Y,
    !,    
    relation_join_1(Ri,X,Y,Ro,Remainder).
relation_join_1([H|Ri],_,_,[],[H|Ri]). % Y1 >@ Y


compose_sh_new(Sin,Seq,So) :-
    rju(Seq,Sin,SeqSin),
    rju(Sin,SeqSin,SinSeqSin),
    ord_union(SeqSin,SinSeqSin,So).

rju(R1,R2,Ro) :-
    rju(R1,R2,R2,XYs,XYs,[],Ro,[]). % using difference list

rju([],_,_,_,_,_,Ro,Ro).

rju([(X,Y)],R2,_,XYs,XYLast,RxzAcc,Front,Tail) :-
%    !,
    relation_join_1(R2,X,Y,R2XY,_),
    XYLast=[(X,Y)],
    ord_union(XYs,R2XY,XYsR2XY),
    ord_union(XYsR2XY,RxzAcc,Rxz),
    append(Rxz,Tail,Front).

rju([(X,Y),(X,Y1)|R1],R2,R2F,XYs,XYLast,RxzAcc,Front,Tail) :-
    !,
    relation_join_1(R2,X,Y,R2XY,R2New),
    ord_union(R2XY,RxzAcc,RxzAccNew),
    XYLast = [(X,Y)|XYLast1],
    rju([(X,Y1)|R1],R2New,R2F,XYs,XYLast1,RxzAccNew,Front,Tail).

rju([(X,Y),(X1,Y1)|R1],R2,R2F,XYs,XYLast,RxzAcc,Front,Tail) :- %X \= Y
    relation_join_1(R2,X,Y,R2XY,_),
    XYLast=[(X,Y)],
    ord_union(XYs,R2XY,XYsR2XY),
    ord_union(XYsR2XY,RxzAcc,Rxz),
    append(Rxz,Front1,Front),
    !,
    rju([(X1,Y1)|R1],R2F,R2F,XYsNew,XYsNew,[],Front1,Tail). % starting new X




/* pair_comb(ASub1,ASub2,ASub3)<==>(ASub1/\ASub2)=ASub3 */

pair_comb(bot,_,bot) :-
    !.
pair_comb(_,bot,bot) :-
    !.
pair_comb(S1,S2,S3) :-
    ord_union(S1,S2,S3).
    

/* pair_bipai(+Comm,+Gi,-Go) abstract interpretation of
   built-ins where Comm has been compiled from a bip
   call. */

%pair_bipai(_, bot, bot)   :- 
%    !.

pair_bipai(copy,S,S)    :-
    !.
pair_bipai(reduce(Vids),Si,So)  :-
    !,
    sh_red_by_gr(Si,Vids,So).
pair_bipai(solve(Eqs), Si, So)  :-
    !,
    pair_solve(Eqs,Si,So).
pair_bipai(reduceCompose(Vids,St), Si, So) :-
    !,
    sh_red_by_gr(Si,Vids,S1),
    compose_sh(S1,St,So).
pair_bipai(remove(S0), Si, So) :-
    !,
    ord_subtract(Si,S0,So).
pair_bipai(compose(St), Si, So) :-
    !,
    compose_sh(Si,St,So).





/* pair_solve(+Sub, +ASubi, -ASubo) -ASubo is an
   abstract substitution on the same domain as +ASubi.-ASubo
   represents a superset of all the substitutions that
   satisfy both constraints +Sub and +ASubi. 

pair_solve(_,bot,bot) :-
    !.
pair_solve(Sub,ASubi,ASubo) :-
    pair_solve_1(Sub,ASubi,ASubo).

pair_solve_1([],ASubi,ASubi).
pair_solve_1([Xid/(Tm,Tvids)|Eqs],ASubi,ASubo) :-
    soln(Xid,Tm,Tvids,ASubi,ASub1),
    pair_solve_1(Eqs,ASub1,ASubo).

    % note that rhs term of an equation is represented by its multiplicity
    % and the set of identifiers of the variables in it

*/

pair_solve(_,bot,bot) :-
    !.
pair_solve(Sub,ASubi,ASubo) :-
    gVars_ngEc(Sub,GVars,Sub1),
    prune(GVars,ASubi,Pruned),
    pair_solve_1(Sub1,Pruned,ASubo).

pair_solve_1([],ASubi,ASubi).
pair_solve_1([Xid/(Tm,Tvids)|Eqs],ASubi,ASubo) :-
    soln(Xid,Tm,Tvids,ASubi,ASub1),
    pair_solve_1(Eqs,ASub1,ASubo).

    % Extract grounded variables and non-grounding equational constraint. 
    gVars_ngEc([],[],[]).
    gVars_ngEc([Xid/(Tm,Tvids)|Eqs],GVars,NGEqs) :- 
         (    Tm ==0 
	  ->  GVars = [Xid|GVars1], gVars_ngEc(Eqs,GVars1,NGEqs)
	  ;   NGEqs = [Xid/(Tm,Tvids)| NGEqs1], gVars_ngEc(Eqs,GVars,NGEqs1)
	 ).

    prune([],ASub,ASub).
    prune([GVar|GVars],ASubi,ASubo) :- 
       prune_1(ASubi,GVar,ASub_),
       prune(GVars,ASub_,ASubo).

    prune_1([],_,[]).
    prune_1([(Xid,Yid)|ASubi],GVar,ASubo) :- 
       ( (Xid=\=GVar,Yid=\=GVar) -> ASubo=[(Xid,Yid)|ASubo_]; ASubo=ASubo_ ),
       prune_1(ASubi,GVar,ASubo_).


soln(Xid,Tm,Tvids,Si,So) :- 
    multiplicity(1,[Xid],Si,Kx),
    multiplicity(Tm,Tvids,Si,Kt),
    unieq_sh(Kx,Kt,Xid,Tvids,Se),
    compose_sh(Si,Se,So). 
    

/* PRIVATE multiplicity(+Tm,+Tvids,+ASub,-C) -C is an approximation of
   the multiplicity of a term represented by (Tm,Tvids) in ASub.  Tm
   and Tvids is pre-computed to avoid repeated computation.  */


multiplicity(0,_,_,0).
multiplicity(2,_,_,2).
multiplicity(1,Tvids,S,C) :- 
    ( 
       	ord_setproduct(Tvids,Tvids,Tvids2), 
       	ord_intersection(S,Tvids2,STvids2), 
       	\+ STvids2 == [] 
    -> 
        C = 2 
    ; 
        C = 1 
    ). 


multi_occ([(_,O)|VOL]) :-
	O >= 2 ; multi_occ(VOL).


/*
sh_red_by_gr([],_,[]).
sh_red_by_gr([(X,Y)|S],Vg,So) :-
    sh_red_by_gr(S,Vg,So1),
    (
        (
            ord_subset([X],Vg) 
	    ;
            ord_subset([Y],Vg)
    	)
    ->
        So = So1
    ;
        So = [(X,Y)|So1]
    ).
*/


sh_red_by_gr([],_,[]).
sh_red_by_gr([(X,Y)|S],Vg,[(X,Y)|So]) :-
    \+ ord_subset([X],Vg),
    \+ ord_subset([Y],Vg),
    !,
    sh_red_by_gr(S,Vg,So).
sh_red_by_gr([XY|S],Vg,[XY|So]) :-
    sh_red_by_gr(S,Vg,So).


/* pair_internalize(+ASubIn,-ASubO) ASubO is obtained
   from ASubIn by replacing each occurrence of a
   variable by its internal Id which is an integer */

pair_internalize(bot,bot).
pair_internalize([],[]).
pair_internalize([(V,V1)|Vs],[(Id,Id1)|Ids]) :-
    variable_id(V,Id),
    variable_id(V1,Id1),
    pair_internalize(Vs,Ids).

pair_externalize(bot,bot).
pair_externalize([],[]).
pair_externalize([(Id,Id1)|Ids],[(V,V1)|Vs]) :-
    variable_id(V,Id),
    variable_id(V1,Id1),
    pair_externalize(Ids,Vs).

pair_tagging(bot,_N,bot).
pair_tagging([],_N,[]).
pair_tagging([(Id,Id1)|Ids],N,[(NId,NId1)|NIds]) :-
    NId is Id+N,
    NId1 is Id1+N,
    pair_tagging(Ids,N,NIds).

vars_internalize([],[]).
vars_internalize([V|Vs],[Id|Ids]) :-
    variable_id(V,Id),
    vars_internalize(Vs,Ids).


/* pair_compile(+Call,-Comm) Compiles Call into Comm. Fails if Call
   definitely fails. Therefore, there is no clause for fail/false and
   so on. */

pair_compile(abolish(_), copy).
pair_compile(format(_,_), copy).
pair_compile(format(_,_,_), copy).
pair_compile(erase(_), copy).
pair_compile(recorda(_,_,_), copy).
pair_compile(recorded(_,_,_), copy).
pair_compile(retract(_), copy).
pair_compile(assert(_), copy).
pair_compile(repeat, copy). 
pair_compile(listing, copy).
pair_compile(listing(_), copy).
pair_compile(compound(_), copy).
pair_compile(portray_clause(_), copy).
pair_compile(\=(_,_), copy).
pair_compile(!, copy).
pair_compile(display(_Term), copy).
pair_compile(nl, copy).
pair_compile(print(_Term), copy).
pair_compile(write(_Term), copy).
pair_compile(writeq(_Term), copy).
pair_compile('@<'(_T1,_T2), copy).
pair_compile('@>'(_T1,_T2), copy).
pair_compile('@=<'(_T1,_T2), copy).
pair_compile('@>='(_T1,_T2), copy).
pair_compile(true, copy).
pair_compile(nonvar(_), copy).
pair_compile(\==(_L,_R), copy).



pair_compile(put(Num), reduce(Vids))  :-
    vars(Num,Vs),
    vars_internalize(Vs,Vids). 
pair_compile(tab(Num), reduce(Vids))  :-
    vars(Num,Vs),
    vars_internalize(Vs,Vids). 
pair_compile(integer(Num), reduce(Vids))  :-
    vars(Num,Vs),
    vars_internalize(Vs,Vids). 
pair_compile(float(Num), reduce(Vids))  :-
    vars(Num,Vs),
    vars_internalize(Vs,Vids). 
pair_compile(number(Num), reduce(Vids))  :-
    vars(Num,Vs),
    vars_internalize(Vs,Vids). 
pair_compile(string(Term), reduce(Vids))  :-
    vars(Term,Vs),
    vars_internalize(Vs,Vids). 
pair_compile(atom(Term), reduce(Vids))  :-
    vars(Term,Vs),
    vars_internalize(Vs,Vids). 
pair_compile(atomic(Term), reduce(Vids))  :-
    vars(Term,Vs),
    vars_internalize(Vs,Vids). 
pair_compile(=:=(N1,N2), reduce(Vids))  :-
    vars((N1,N2),Vs),
    vars_internalize(Vs,Vids). 
pair_compile(=\=(N1,N2), reduce(Vids))  :-
    vars((N1,N2),Vs),
    vars_internalize(Vs,Vids). 
pair_compile(<(N1,N2), reduce(Vids))  :-
    vars((N1,N2),Vs),
    vars_internalize(Vs,Vids). 
pair_compile(>(N1,N2), reduce(Vids))  :-
    vars((N1,N2),Vs),
    vars_internalize(Vs,Vids). 
pair_compile(=<(N1,N2), reduce(Vids))  :-
    vars((N1,N2),Vs),
    vars_internalize(Vs,Vids). 
pair_compile(>=(N1,N2), reduce(Vids))  :-
    vars((N1,N2),Vs),
    vars_internalize(Vs,Vids). 
pair_compile(length(_,Num), reduce(Vids))  :-
    vars(Num,Vs),
    vars_internalize(Vs,Vids). 
pair_compile(ground(T), reduce(Vids))  :-
    vars(T,Vs),
    vars_internalize(Vs,Vids). 
pair_compile(statistics(K,S), reduce(Vids))  :-
    vars((K,S),Vs),
    vars_internalize(Vs,Vids). 
pair_compile(functor(_,PN,PA), reduce(Vids))  :-
    vars((PN,PA),Vs),
    vars_internalize(Vs,Vids). 
pair_compile(name(A,L), reduce(Vids))  :-
    vars((A,L),Vs),
    vars_internalize(Vs,Vids). 
pair_compile(compare(O,_,_), reduce(Vids))  :-
    vars(O,Vs),
    vars_internalize(Vs,Vids). 
pair_compile(is(L,R), reduce(Vids))  :-
    vars((L,R),Vs),
    vars_internalize(Vs,Vids). 

pair_compile(=@=(L,R), solve(Eqs))  :-
    mgu(L,R,E0),
    pair_eq_constraints(E0,Eqs).
pair_compile(=(L,R), solve(Eqs))  :-
    mgu(L,R,E0),
    pair_eq_constraints(E0,Eqs).
pair_compile(==(L,R), solve(Eqs))  :-
    mgu(L,R,E0),
    pair_eq_constraints(E0,Eqs).
pair_compile(=..(L,R), solve(Eqs))  :-
    mgu(L,R,E0),
    pair_eq_constraints(E0,Eqs).
pair_compile(sort(L,R), solve(Eqs))  :-
    mgu(L,R,E0),
    pair_eq_constraints(E0,Eqs).
pair_compile(keysort(L,R), solve(Eqs))  :-
    mgu(L,R,E0),
    pair_eq_constraints(E0,Eqs).

pair_compile(arg(Arg,Term,Value), reduceCompose(Vids,St)) :-
    vars(Arg,Va),
    vars_internalize(Va,Vids),
    vars((Term,Value),Vars),
    vars_internalize(Vars,Vids1),
    ord_setproduct(Vids1,Vids1,St).

pair_compile(var(V), remove([(Vid,Vid)])) :-
    variable(V),
    variable_id(V,Vid).

pair_compile(read(Term), compose(St)) :-
    /* reading a term from a device and unifying it with 
       Term may cause sharing between variables in Term and those 
       sharing with them */
    vars(Term,Vars),
    vars_internalize(Vars,Vids),
    ord_setproduct(Vids,Vids,St).


/* pair_eq_constraints(+Sun,-Eqs) Each equation of the form X/T is
   transformed into an equational constraint of the form
   Xid/(Tm,Tvids) where Xid is the identifier of X and Tm is 2 if T is
   non-linear, 0 if T is ground or 1 otherwise. Tvids is the set of
   identifiers of the variables in T.  */


pair_eq_constraints([],[]).
pair_eq_constraints([X/T|Sub],[Xid/(Tm,Tvids)|Eqs]) :- 
    variable_id(X,Xid),
    vars(T,Vt),
    vars_internalize(Vt,Tvids),
    var_occ(T,VCL),
    ( multi_occ(VCL) -> Tm =2; \+ (Vt == []) -> Tm=1; Tm=0 ),
    pair_eq_constraints(Sub,Eqs).


/*
test(W) :- 
    left_right(5,[1,2,3,4,6,7,8,9],W).
*/

/**************************************************************************
 The following code is for backward analysis
**************************************************************************/

/* powset(Si,So) So is the powset of Si */ 

/*

powset([],[[]]).
powset([X|Xs],Zs) :- 
    powset(Xs,Ys),
    prepend(Ys,X,Ys,Zs).

prepend([],_,Acc,Acc).
prepend([S|Xs],X,Acc,[[X|S]|Ys]) :- 
    prepend(Xs,X,Acc,Ys). 

*/




/* subset_gen(+Si,-So) So is a subset of Si. Enumerate all subsets on
    backtracking. The implementation uses partial structures to hold
    elements which is alread picked to be in the subset. This
    implementation satisfies an important property that subsets are
    enumerated in reverse topological order with respect to setset
    relation.  


subset_gen(Si,So) :- subset_gen(Si,So,So). 

subset_gen([],_,[]). % The hole is filled with empty list
subset_gen([X|Xs],Front,Tail) :- 
    Tail = [X|Tail1],
    subset_gen(Xs,Front,Tail1).
subset_gen([_|Xs],Front,Tail) :- 
    subset_gen(Xs,Front,Tail).

*/

/* sub_pair_sharing(+Si,-So) So is a symmetric subset of Si. Enumerate
    all such subsets on backtracking. The implementation uses partial
    structures to hold elements which is alread picked to be in the
    subset. This implementation satisfies an important property that
    subsets are enumerated in reverse topological order with respect
    to setset relation.  */

sub_pair_sharing(Si,So) :- sub_pair_sharing(Si,So,So). 

sub_pair_sharing([],_,[]). % The hole is filled with empty list
sub_pair_sharing([(X,Y)|Ps],Front,Tail) :- 
    X @=< Y, % choose to pick (X,Y) and (Y,X)
    Tail = [(X,Y)|Tail1],
    sub_pair_sharing(Ps,Front,Tail1).
sub_pair_sharing([(X,Y)|Ps],Front,Tail) :-
    X @=< Y, % choose not to pick (X,Y), (Y,X)
    sub_pair_sharing(Ps,Front,Tail).
sub_pair_sharing([(Y,X)|Ps],Front,Tail) :- 
    Y @> X, 
    ( 
         is_in(X,Y,Front)
    -> 
         Tail = [(Y,X)|Tail1],
         sub_pair_sharing(Ps,Front,Tail1)
    ;
         sub_pair_sharing(Ps,Front,Tail)
    ).

is_in(_,_,Empty) :- 
    var(Empty),
    !,
    fail.
is_in(X,Y,[(X,Y)|_]) :- !.
is_in(X,Y,[_|List]) :- is_in(X,Y,List).

/*test:-
    tell('sub_pair_sharing.log'),
    sub_pair_sharing(
      [(a,a),(a,b),(a,c),(b,a),(b,b),(b,c),(c,a),(c,b),(c,c)],L),
    write(L),nl,fail
    ; told.
*/


/*pair_solve_abduce(E, Post, Pre) :- 
    sub_pair_sharing(Post,Pre),
    pair_solve(E,Pre,TightPost),
    ord_subset(TightPost,Post).
*/

/** an improvement over this is to make use of monotonicity of
    pair_solve -- the function to be abduced. Starting from Post, then
    maximal subsets of Post and then maximal subsets of those maximal
    subsets. If a subset of Post is a valid pre-condition for the
    post-condition Post, no subsets of that subset need be
    considered. Of course, the best way to approach this is to find a
    constructive method that avoids applying pair_solve 

pair_solve_abduce(E, Post, Pre) :- 
    retractall(in_database(_)),
    sub_pair_sharing(Post,Pre),
    \+ is_in_database(Pre),
    write(testing(Pre)),nl,
    pair_solve(E,Pre,TightPost),
    ord_subset(TightPost,Post),
    assert(in_database(Pre)).


:- dynamic in_database/1. 

is_in_database(X) :- 
    in_database(Y),
    ord_subset(X,Y).

*/

/* Implementation of new algorithm */

/* Operator {X} \otimes Ys = ({X} \times Ys) \cup (Ys \times {X}) -
   assuming that Ys is oredered set. */

otimes(X,Ys,XYsYsX) :- 
    otimes_prod(Ys,X,XYs,YsX),
    ord_union(XYs,YsX,XYsYsX).

    otimes_prod([],_,[],[]).
    otimes_prod([Y|Ys],X,[(X,Y)|XYs],[(Y,X)|YsX]) :- otimes_prod(Ys,X,XYs,YsX).

otimes_test :- 
    Ys = [1,2,3],
    otimes(2,Ys,Ys33Ys),
    write(Ys33Ys),nl.

/* Function svars(+Pi,+X,-PiX) PiX is the set of variables that share
    with X under Pi. Pi is a symmetric ordered set of pairs. */

svars([],_,[]).
svars([(Y,X)|Pi],X,[Y|PiX]) :- 
    !, 
    svars(Pi,X,PiX). 
svars([_|Pi],X,PiX)         :- 
    svars(Pi,X,PiX).

svars_test :- 
    Pi = [(1,2),(1,3),(3,3),(2,1),(3,1)],
    svars(Pi,3,Pi3),
    write(Pi3),nl.

/* Operator Phi \uplus Psi = { phi \cup psi | phi\in Phi and psi \in
   Psi}. The implementation uses a difference list. */

uplus(Sxs,Sys,Szs) :- uplus(Sxs,Sys,Szs,[]).

uplus([],_,Szs,Szs).
uplus([Sx|Sxs],Sys,SzsH,SzsT) :- 
    uplus_1(Sys,Sx,SzsH,SzsH1),
    uplus(Sxs,Sys,SzsH1,SzsT).

uplus_1([],_,SzsH,SzsH). 
uplus_1([Sy|Sys],Sx,[Sz|SzsH],SzsT) :- 
    ord_union(Sx,Sy,Sz),
    uplus_1(Sys,Sx,SzsH,SzsT). 

uplus_test :- 
    Sxs = [[1,2],[1,3]],
    Sys = [[a,b],[a,c],[b,c]], 
    uplus(Sxs,Sys,Szs),
    write(Sxs),nl,
    write(Sys),nl,
    write(Szs),nl.

vectorise(NgEC,(1,LVars),Right) :- 
    vectorise(NgEC,LVars,(1,[]),Right).

vectorise([],[],Right,Right).
vectorise([Xid/(Tm,TVids)|EC],[Xid|LVars],(RM,RVars),Right) :-
    ord_union(TVids,RVars,RVars1),
    ( RM=2->RM1=2; Tm=2->RM1=2; ord_intersection(TVids,RVars,[])->RM1=1; RM1=2 ),
    vectorise(EC,LVars,(RM1,RVars1),Right).

fuses(U,V,(Sm,SVars),(Tm,TVars),Pi,Fs) :-
    ( member(U,SVars) -> UinS = yes ; UinS = no ),
    ( member(V,SVars) -> VinS = yes ; VinS = no ),
    ( member(U,TVars) -> UinT = yes ; UinT = no ),
    ( member(V,TVars) -> VinT = yes ; VinT = no ),
    fuses_1(UinS,VinT,U,V,SVars,TVars,Fs1), 
    fuses_1(UinT,VinS,U,V,TVars,SVars,Fs4), 
    fuses_2(Tm,UinS,VinS,U,V,SVars,Fs2),
    fuses_2(Sm,UinT,VinT,U,V,TVars,Fs5),
    fuses_3(UinS,VinS,U,V,SVars,TVars,Pi,Fs3),
    fuses_3(UinT,VinT,U,V,TVars,SVars,Pi,Fs6),
    uplus(Fs1,Fs2,Fs12),
    uplus(Fs3,Fs12,Fs123),
    uplus(Fs4,Fs123,Fs1234),
    uplus(Fs5,Fs1234,Fs12345),
    uplus(Fs6,Fs12345,Fs).

fuses_1(yes,yes,_U,_V,_SVars,_TVars,[]).
fuses_1(yes,no, _U, V,_SVars, TVars,[VTVars]) :- 
    otimes(V,TVars,VTVars).
fuses_1(no, yes, U,_V, SVars,_TVars,[USVars]) :- 
    otimes(U,SVars,USVars).
fuses_1(no, no,  U, V, SVars, TVars,[VTVars, USVars]) :- 
    otimes(V,TVars,VTVars),
    otimes(U,SVars,USVars).

fuses_2(1,_UinS,_VinS,_U,_V,_SVars,[[]]). 
fuses_2(2, UinS, VinS, U, V, SVars,Fs)  :- 
    fuses_2_nonlinear(UinS,VinS,U,V,SVars,Fs).

    fuses_2_nonlinear(yes,yes,_U,_V,_SVars,[]).
    fuses_2_nonlinear(yes,no, _U, V, SVars,[VSVars]) :- 
          otimes(V,SVars,VSVars).
    fuses_2_nonlinear(no, yes, U,_V, SVars,[USVars]) :- 
          otimes(U,SVars,USVars).
    fuses_2_nonlinear(no, no,  U, V, SVars,[USVars,VSVars]) :- 
          otimes(U,SVars,USVars),
          otimes(V,SVars,VSVars).

fuses_3(yes,yes,_U,_V,_SVars,TVars,Pi,[PiT]) :- 
    ord_setproduct(TVars,TVars,TVars2),
    ord_intersection(TVars2,Pi,PiT).
fuses_3(yes,no, _U, V, SVars,TVars,Pi,[PiT,VSVars]) :- 
    ord_setproduct(TVars,TVars,TVars2),
    ord_intersection(TVars2,Pi,PiT),
    otimes(V,SVars,VSVars).
fuses_3(no, yes, U,_V, SVars,TVars,Pi,[PiT,USVars]) :- 
    ord_setproduct(TVars,TVars,TVars2),
    ord_intersection(TVars2,Pi,PiT),
    otimes(U,SVars,USVars).
fuses_3(no, no,  U, V, SVars,TVars,Pi,[PiT,USVars,VSVars]) :- 
    ord_setproduct(TVars,TVars,TVars2),
    ord_intersection(TVars2,Pi,PiT),
    otimes(U,SVars,USVars),
    otimes(V,SVars,VSVars).


fuses_all(Delta,S,T,Pi,Fs) :- 
    fuses_all(Delta,S,T,Pi,[[]],Fs).

fuses_all([],_,_,_,Fs,Fs). % uses an accumulator 
fuses_all([(U,V)|Delta],S,T,Pi,FsAcc,Fs) :- 
    fuses(U,V,S,T,Pi,FsUV),
    uplus(FsUV,FsAcc,FsAcc1),
    fuses_all(Delta,S,T,Pi,FsAcc1,Fs). 
    
/** This simplified implementation assumes that E does not have grounding equations. 
*/

pair_solve_abduce(E, Post, Pre) :- 
    pair_solve(E, Post, Post1),
    ord_subtract(Post1,Post,Delta1),
    remove_symmetry(Delta1,Delta),
    vectorise(E,S,T), 
    fuses_all(Delta,S,T,Post,Fs),
    write('Fuses:'), nl, write_list(Fs),nl,nl,
    minimize(Fs,FsMin),
    write('Minimal Fuses:'), nl, write_list(FsMin),nl,nl,
    member(F,FsMin),
    ord_subtract(Post,F,Pre),
    ( 
      pair_solve(E,Pre,Post2),
      ord_subset(Post2,Post)
    ->
        write('Verified'),nl
      ; write('Oops!'),nl
    ).

    remove_symmetry([],[]).
    remove_symmetry([(X,Y)|Pi],[(X,Y)|Phi]) :-
         X =< Y,
         !,
         remove_symmetry(Pi,Phi).
    remove_symmetry([_|Pi], Phi) :-
         % X > Y,
         !,
         remove_symmetry(Pi,Phi).

minimize(Fsi,Fso) :- minimize(Fsi,Fsi,Fso).

    minimize([],_,[]).
    minimize([F|Fsi],Fs,Fso) :- 
          subsumed(Fs,F) 
      ->  minimize(Fsi,Fs,Fso) 
      ;   Fso = [F|Fso1], minimize(Fsi,Fs,Fso1). 

subsumed([F1|Fs],F) :- 
    (\+ F = F1, ord_subset(F1,F)) -> true ; subsumed(Fs,F).

    write_list([])     :- nl.
    write_list([X|Xs]) :- write(X),nl,write_list(Xs). 


%%%

transitive(R,C) :-
    relation_join(R,R,R2),
    ( 
       ord_subset(R2,R)
    -> 
       C = R
      ;
       ord_union(R,R2,R3),
       transitive(R3,C)
    ).

test :- transitive([(1,2),(1,5),(2,4),(3,2),(4,3),(5,4)],C),write(C),nl.
