/* 
   This program is used to generate a decision program for conjunctions 
   of primitive literals for a specific set of type primitives. You 
   need to 
     (1) modify primitive_types/1.
     (2) less_than/2
     (3) pcode/1 and ncode/1 
   and run generate/0. 

   You get a table $etype_conj_prim_coded/1 facts that decides if a conjunction 
   of primitive literals that is coded by a unique number is empty.  
   The run-time program will consists of 
   (1) $etype_conj_prim_coded plus
   (2) coding/3, ncoding/2,pcoding/2,pcode/1 and ncode/1 
   (3) etype_conj_prim/2

   generate(+TCS) outputs all possible conjunctive types constructible
   from the set of type constant TCS. A conjunctive type is
   represented by a pair Ps,Ns such that Ps and Ns are subsets of TCS
   and Ps and Ns are non-overlapping. */

:- use_module(ordsets). 

%:- use_module(library(lists)).


generate :- 
    primitive_types(TCS),
    sort(TCS,TCS0),
    sublist(TCS0,TCS1),
    sublist(TCS1,Ps),
    ord_subtract(TCS1,Ps,Ns),
    etype_conj_prim_gen(Ps,Ns),
    coding(Ps,Ns,Num),
    writeq('$etype_conj_prim_coded'(Num)), write('.'), 
    write('%'), write(empty(Ps,Ns)), nl,
    fail.
    

sublist([],[]).
sublist([H|L],[H|S]) :- sublist(L,S).
sublist([_|L],S) :- sublist(L,S).
    
etype_conj_prim_gen(Ps,Ns) :-  
    % can be replaced by coding(Ps,Ns,Num), '$etype_coded'(Num)
    % provided generated facts for '$etype_coded'/1 are all included
    simplify_conj_pos(Ps,Ps0), 
    (
       Ps0 = [_,_|_] % Ps0 containing non-overlapping primitive types
     ;
    
       Ps0 = [P],
       filter(Ns,P,Ns0), 
       simplify_conj_neg(Ns0,Ns1),                                
       frontier(P,Ns1)
    ).

/* simplify_conj_pos(+Pos1,-Pos2). Pos2 is obtained by removing any 
   primitive type from Pos1 that is less than some other in Pos1. */

simplify_conj_pos(PTsi,PTso) :- 
    simplify_conj_pos(PTsi,[],PTso).


simplify_conj_pos([],PTsu,PTsu).
simplify_conj_pos([PTi|PTsi],PTsu, PTso) :- 
    append(PTsu,PTsi,PTsn),
    (
        member(PT,PTsn),less_than(PT,PTi)
    -> 
        simplify_conj_pos(PTsi,PTsu,PTso)
    ;
        simplify_conj_pos(PTsi,[PTi|PTsu],PTso)
    ).

/* simplify_conj_neg(+Neg1,-Neg2). Neg2 is obtained by removing any 
   primitive type from Neg1 that is greater than some other in Neg1. */

simplify_conj_neg(PTsi,PTso) :- 
    simplify_conj_neg(PTsi,[],PTso).


simplify_conj_neg([],PTsu,PTsu).
simplify_conj_neg([PTi|PTsi],PTsu, PTso) :- 
    append(PTsu,PTsi,PTsn),
    (
        member(PT,PTsn),less_than(PTi,PT)
    -> 
        simplify_conj_neg(PTsi,PTsu,PTso)
    ;
        simplify_conj_neg(PTsi,[PTi|PTsu],PTso)
    ).

/* filter(+Nsi,+P,-Nso) 
     Let Ns is in Nsi. If less_than(P,Ns), P is in Nso;
                       If less_than(Ns,P) or Ns = P, Ns is in Nso; 
                       Otherwise, Ns is not in P. 
*/

filter([],_,[]).
filter([Ni|Nsi],P,Nso) :- 
    ( 
        ( P=Ni ; less_than(Ni,P) )
    -> 
        Nso = [Ni|Nso1], 
        filter(Nsi,P,Nso1)
     ;
        less_than(P,Ni) 
    ->  
        Nso = [P|Nso1],
        filter(Nsi,P,Nso1) 
     ;
        filter(Nsi,P,Nso)
    ). 




%%%%%%%%%%%%%%%%%%%%%%%%%%specific_type_system %%%%%%%%%%%%%%%%%%%%%%%%%%%%

/*
   The following is input to generate/0. This tool requires that
   primitives satisfy the following condition. (1) A primitive is the
   union of all of its subordinate primitives; and (2) two different
   primitive types overlap iff one is a subtype of another. The
   following lattice of primitives satisfy this condition.

                  
                        atomic
                      /         \
                     atom         number
                                  /   \
                                 float integer

   You may need to add dummy types so as to satisfy (1)  as in the
   following.  You may need to add more than one dummies.


%                   ground
%                   /  \
%                  /    \
%                 /      \
%              other     atomic
%              ^^^^^   /        \
%                     atom       number
%                                  /   \
%                                 float integer

   The conditions (1) and (2) being satisfied, a simplified
   conjunction of primitive literals equals to empty if its positive
   part is not singleton. Thus, we only need to consider the case
   where the positive part is a singleton. In this case, P and ~N1 and
   ~N2 and ... and ~Nk is empty iff {N1,N2,...Nl} is a frontier of the
   tree rooted at P where {N1,N2,...Nl} are obtained by replacing
   those in {N1,...,Nk} that are greater than P with P and retaining
   those that are small than P and removing others.

   frontier(+P,+Ns) Ns is a frontier of P. When dummy types are added,
   frontier only lists those frontiers that do not contain dummies. 

   We code each atom with two bits, one for it appearing in positive 
   part and the other for it appearing in negative part.

*/

% do not put dummy primitive types

primitive_types([atom,atomic,float,integer,number,string]). % primitive types

% no need to list less_than for dummies.

less_than(atom,atomic).    % an atom is atomic
                           % atomic is no less than primitive type
less_than(float,number).   % a float is a number and atomic
less_than(float,atomic). 
less_than(integer,number). % a float is a number and atomic
less_than(integer,atomic). 
less_than(number,atomic).  % a number is atomic


frontier(atomic,[atomic]).
frontier(atomic,[atom,number]).
frontier(atomic,[atom,float,integer]).
frontier(atom,[atom]). 
frontier(number,[number]).
frontier(number,[float,integer]).
frontier(float,[float]).
frontier(integer,[integer]). 


% again no need to list dummies.

pcode(atom,   1).    % code for an atom in positive part
pcode(atomic, 2).
pcode(float,  4).
pcode(integer,8).
pcode(number,16).

ncode(atom,     512). % code for an atom in negative part
ncode(atomic,  1024).
ncode(float,   2048).
ncode(integer, 4096).
ncode(number,  8192).


etype_conj_prim(Ps,Ns) :-  
    % provided that all generated facts for '$etype_conj_prim_coded'/1 are included
    coding(Ps,Ns,Num), 
    '$etype_conj_prim_coded'(Num).

/* coding(+Ps,+Ns,Num) returns a unique code for a conjunction of
   primitive type literals */

coding(Ps,Ns,Num) :- 
   pcoding(Ps,PNum),
   ncoding(Ns,NNum),
   Num is PNum+NNum. 

pcoding(Ps,Num) :- 
   pcoding(Ps,0,Num).

   pcoding([],Num,Num).
   pcoding([P|Ps],Num0,Num) :- 
      pcode(P,Bit),
      Num1 is Num0+Bit,
      pcoding(Ps,Num1,Num).

ncoding(Ps,Num) :- 
   ncoding(Ps,0,Num).

   ncoding([],Num,Num).
   ncoding([P|Ps],Num0,Num) :- 
      ncode(P,Bit),
      Num1 is Num0+Bit,
      ncoding(Ps,Num1,Num).









