/* 

An abstract substitution on a set of variables Vars is represented as
a set of variable-typings whose domain is a subset of Vars. A
variable-typing maps a variable not in its domain to the type
"top". The choice of a partial function instead of a total function is
non-essential.

*/

:- module(rtype, [ 
		rtype_eq_constraints/2,
                rtype_init/0,
		rtype_list_print/1,
		rtype_unify_entry/5,
		rtype_unify_exit/6,
		rtype_bipai/3,
		rtype_compile/2,
		rtype_bot/2,
		rtype_ident/2,
		rtype_join/3,
		rtype_leq/2,
		rtype_internalize/2,
		rtype_externalize/2,
		rtype_bw_compile/2,
		rtype_bip_has_cond/2,
		rtype_top/2,
		rtype_meet/3,
		rtype_bw_unify_entry/5,
		rtype_bw_unify_exit/6,
		rtype_bw_bipai/3
	       ]).


:- use_module(basics, [
	mgu/3,
        variable/1,
	vars/2
   ]).


:- use_module(typedef, [
	typedef_constructors/1,
	typedef_functors/1,
	typedef/3
   ]).
     
:- use_module(counter).

:- use_module(sicstus). 

:- [empty].


current_time(MiniSec) :-  statistics(runtime,[MiniSec,_]).


/* rtype_eq_constraints(+Sub,-Eqs) - do nothing */

rtype_eq_constraints(Sub,Sub).


rtype_externalize(vtbot,[]).
rtype_externalize([],[[]]).
rtype_externalize([H|L],VTS) :- 
    ( vt_bot([H|L]) -> VTS = [] ; VTS=[[H|L]] ).
rtype_externalize(or(DVTS1,DVTS2),VTS) :- 
    rtype_externalize(DVTS1,VTS1),
    rtype_externalize(DVTS2,VTS2), 
    append(VTS1,VTS2,VTS).

rtype_internalize([],vtbot).
rtype_internalize([VT],VT).
rtype_internalize([VT1,VT2|VTS],or(VT1,DVTS)) :- 
    rtype_internalize([VT2|VTS],DVTS).



/* The lower bound of two sets of variable typings. This is different
   from the combination of two abstract substitution because the
   former works on two sets of variable typings on the overlapping
   domains while the later on disjoint domains. */


vtsglb(vtbot,_,vtbot).
vtsglb(or(DVTS11,DVTS12),DVTS2,DVTS) :- 
    vtsglb(DVTS11,DVTS2,DVTSo1),
    (
        DVTSo1 == vtbot
    -> 
        vtsglb(DVTS12,DVTS2,DVTS)
    ;
        vtsglb(DVTS12,DVTS2,DVTSo2),
        (
             DVTSo2 == vtbot 
        ->
             DVTS = DVTSo1
        ;
             DVTS = or(DVTSo1,DVTSo2)
        )
    ).

vtsglb([],DVTS,DVTS). % [] is the top variable typing
vtsglb([H|L],DVTSi, DVTSo) :- 
    vtsglb_1(DVTSi,[H|L],DVTSo). 

vtsglb_1(vtbot,_,vtbot).
vtsglb_1(or(DVTS1,DVTS2),VT,DVTS) :- 
    vtsglb_1(DVTS1,VT,DVTSo1),
    (
        DVTSo1 == vtbot
    -> 
        vtsglb_1(DVTS2,VT,DVTS)
    ;
        vtsglb_1(DVTS2,VT,DVTSo2),
        (
             DVTSo2 == vtbot 
        ->
             DVTS = DVTSo1
        ;
             DVTS = or(DVTSo1,DVTSo2)
        )
    ).
vtsglb_1([],VT,VT). % [] is the top variable typing
vtsglb_1([H|L],VT, VTo) :- 
    vtglb([H|L],VT,VTo). 


/* vtglb(+VT1,+VT2,-VT3) +VT1 and +VT2 are variable-typings. vtglb/3
   succeeds with dom(VT3)=dom(VT1)+dom(VT2) and for each varaible X in
   dom(VT3) such that X/T is in VT1 and X/T2 in VT2, X/T3 is in VT3
   and T3 is the greatest lower bound of T1 and T2.

   This implementation assumes that variable typings are sorted lists
   that are sorted according to the ordering on the names of
   variables.*/

/*
vtglb([],VT,VT) 	:-
        !.
vtglb(VT,[],VT) 	:-
        !.
vtglb([X1/T1|VT1],[X2/T2|VT2],[X1/T3|VT3]) 	:-
    	X1 == X2,
    	!,
    	tand(T1,T2,T3),
    	vtglb(VT1,VT2,VT3).
vtglb([X1/T1|VT1],[X2/T2|VT2],[X1/T1|VT3]) 	:-
    	X1 @< X2,
    	!,
    	vtglb(VT1,[X2/T2|VT2],VT3).
vtglb([X1/T1|VT1],[X2/T2|VT2],[X2/T2|VT3]) 	:-
    	X1 @> X2,
    	!,
    	vtglb([X1/T1|VT1],VT2,VT3).
*/


vtglb([],VT,VT).
vtglb([X1/T1|VT1],VT2,VT3) :- 
    vtglb_1(VT2,X1,T1,VT1,VT3).

vtglb_1([],X1,T1,VT1,[X1/T1|VT1]).
vtglb_1([X2/T2|VT2],X1,T1,VT1,[X3/T3|VT3]) :- 
    (  	X1 == X2
     ->
        X3=X1,
    	tand(T1,T2,T3),
    	vtglb(VT1,VT2,VT3)
     ;
    	X1 @< X2
     ->
        X3=X1, T3=T1,
    	vtglb_1(VT1,X2,T2,VT2,VT3)
     ;
        X3=X2,T3=T2, % X1 @> X2
    	vtglb_1(VT2,X1,T1,VT1,VT3)
     ).




rtype_list_print(_).


rtype_unify_entry(Sub,Sigma,_NVarR,NVar,Zeta) :- 
           vts_tag(Sigma,NVar,SigmaR), 
	   rtype_solve(Sub,SigmaR,SigmaRSub),
           vts_restrict(SigmaRSub,NVar,Zeta).

    vts_tag(vtbot,_Shift,vtbot).
    vts_tag([],_Shift,[]).
    vts_tag(['$VAR'(Ni)/Ti|VTi],Shift,['$VAR'(No)/Ti|VTo]) :-
	    No is Ni + Shift,
	    vts_tag(VTi,Shift,VTo).
    vts_tag(or(DVTS1,DVTS2),Shift,or(DVTSo1,DVTSo2)) :-
        vts_tag(DVTS1,Shift,DVTSo1),
        vts_tag(DVTS2,Shift,DVTSo2).


    vts_restrict(vtbot,_NVar,vtbot).
    vts_restrict([],_NVar,[]).
    vts_restrict(['$VAR'(Ni)/Ti|VTi],NVar,VTo) :-
	    ( Ni =< NVar -> VTo = ['$VAR'(Ni)/Ti|VT] ; VTo=VT ), 
	    vts_restrict(VTi,NVar,VT).
    vts_restrict(or(DVTS1,DVTS2),NVar,or(DVTSo1,DVTSo2)) :-
        vts_restrict(DVTS1,NVar,DVTSo1),
        vts_restrict(DVTS2,NVar,DVTSo2).



rtype_unify_exit(Sub,Sigma,_NVarR,Theta,NVar,Zeta) :-
           vts_tag(Sigma,NVar,SigmaR), 
           rtype_comb(SigmaR,Theta,SigmaRTheta),
	   rtype_solve(Sub,SigmaRTheta,SigmaRThetaSub),
%           vts_restrict(SigmaRThetaSub,NVar,Zeta).
           vts_restrict(SigmaRThetaSub,NVar,Zeta1),
           remove_subsumed_vt(Zeta1,Zeta).


rtype_meet(Zeta,Eta,Beta) :-  
    vtsglb(Zeta,Eta,Beta1), 
    reduce(Beta1,Beta).



/*rtype_bw_unify_exit(_CE,_Answer,_NVarR,DemandRt,_NVar,DemandRt). 
% just pass the result to the left
*/


rtype_bw_unify_exit(CE,Answer,_NVarR,DemandRt,NVar,DemandLt) :-
    vts_tag(Answer,NVar,AnswerR),
    rtype_solve(CE,AnswerR,CEAnswerR), 
    vts_restrict(CEAnswerR,NVar,CEAnswerR1),
    pseduo_complement(CEAnswerR1,DemandRt,DemandLt).


/*
rtype_bw_unify_exit(CE,Answer,_NVarR,DemandRt,NVar,DemandLt) :-
	  vts_tag(Answer,NVar,AnswerR),
          vts_tag(DemandRt,NVar,DemandRtR),
          pseduo_complement(AnswerR,DemandRtR,DemandAfterCE),
          rtype_solve_bw(CE,DemandAfterCE,DemandLtUnRestricted), 
          vts_proj_forall(DemandLtUnRestricted,NVar,DemandLt).
*/


rtype_bw_unify_entry(CE,Sigma,_NVarR,NVar,Eta) :-
    vts_tag(Sigma,NVar,SigmaR),
    rtype_solve_bw(CE,SigmaR,Zeta),
    vts_proj_forall(Zeta,NVar,Eta).




/* 

The combination of two abstract substitution is formed by conjucting
two variable-typings each from one abstract substitution in every
possibility.

*/

rtype_comb(vtbot,_,vtbot).
rtype_comb(or(DVTS11,DVTS12),DVTS2,DVTS) :- 
    rtype_comb(DVTS11,DVTS2,DVTSo1),
    (
        DVTSo1 == vtbot
    -> 
        rtype_comb(DVTS12,DVTS2,DVTS)
    ;
        rtype_comb(DVTS12,DVTS2,DVTSo2),
        (
             DVTSo2 == vtbot 
        ->
             DVTS = DVTSo1
        ;
             DVTS = or(DVTSo1,DVTSo2)
        )
    ).

rtype_comb([],DVTS,DVTS). % [] is the top variable typing
rtype_comb([H|L],DVTSi, DVTSo) :- 
    rtype_comb_1(DVTSi,[H|L],DVTSo). 

rtype_comb_1(vtbot,_,vtbot).
rtype_comb_1(or(DVTS1,DVTS2),VT,DVTS) :- 
    rtype_comb_1(DVTS1,VT,DVTSo1),
    (
        DVTSo1 == vtbot
    -> 
        rtype_comb_1(DVTS2,VT,DVTS)
    ;
        rtype_comb_1(DVTS2,VT,DVTSo2),
        (
             DVTSo2 == vtbot 
        ->
             DVTS = DVTSo1
        ;
             DVTS = or(DVTSo1,DVTSo2)
        )
    ).
rtype_comb_1([],VT,VT). % [] is the top variable typing
rtype_comb_1([H|L],VT, VTo) :- 
    append([H|L],VT,VTo). 



/*

The identity abstract substitution on Vars constains a single variable
typing that maps every variable into "top".

*/

rtype_ident(_Vars,[]). 

/* 

The top abstract substitution on Vars constains a single variable
typing that maps every variable into "top".

*/

rtype_top(_Vars,[]).



/* 

The bottom abstract substitution on Vars is the empty disjunction of variable
typings. 

*/

rtype_bot(_Vars, vtbot). 



/* 

rtype_solve(+Sub, +ASubi, -ASubo) -ASubo is an abstract substitution
on the same domain as +ASubi.-ASubo approximates the set of
substitutions that satisfy both constraints +Sub and +ASubi.

rtype_solve(EQS,ASubi,ASubo) :-
	rtype_down(EQS,ASubi,ASubd),
	rtype_up(EQS,ASubd,ASubu),
        remove_vt_bot(ASubu,ASubo).

*/


rtype_solve(EQS,ASubi,ASubo) :-
	rtype_down(EQS,ASubi,ASubd),
	rtype_up(EQS,ASubd,ASubo)
%        ,write(rtype_solve(EQS,ASubi,ASubo)),nl
        .

rtype_down([],ASub,ASub).
rtype_down([LHS/RHS|EQS],ASubi,ASubo) :- 
        rtype_down_1(ASubi,LHS,RHS,ASub1),
        rtype_down(EQS,ASub1,ASubo). 
        

rtype_down_1(vtbot,_,_,vtbot).
rtype_down_1(or(DVTS1,DVTS2),LHS,RHS,DVTS) :- 
    rtype_down_1(DVTS1,LHS,RHS,DVTSo1),
    (
        DVTSo1 == vtbot
    -> 
        rtype_down_1(DVTS2,LHS,RHS,DVTS)
    ;
        rtype_down_1(DVTS2,LHS,RHS,DVTSo2),
        (
             DVTSo2 == vtbot 
        ->
             DVTS = DVTSo1
        ;
             DVTS = or(DVTSo1,DVTSo2)
        )
    ).
rtype_down_1([],_,_,[]). % top variable typing
rtype_down_1([H|L],LHS,RHS,DVTSo) :-
        ( 
            mapsto(LHS,[H|L],Tp), 
            \+ Tp = top
        -> 
	    vts_up_v2(Tp,RHS,[H|L],DVTSo)
	 ;
            DVTSo=[H|L]
        ).



rtype_up([],ASub,ASub).
rtype_up([LHS/RHS|EQS],ASubi,ASubo) :- 
        rtype_up_1(ASubi,LHS,RHS,ASub1),
        rtype_up(EQS,ASub1,ASubo). 

rtype_up_1(vtbot,_,_,vtbot).
rtype_up_1(or(DVTS1,DVTS2),LHS,RHS,DVTS) :-
    rtype_up_1(DVTS1,LHS,RHS,DVTSo1),
    (
        DVTSo1 == vtbot
    -> 
        rtype_up_1(DVTS2,LHS,RHS,DVTS)
    ;
        rtype_up_1(DVTS2,LHS,RHS,DVTSo2),
        (
             DVTSo2 == vtbot 
        ->
             DVTS = DVTSo1
        ;
             DVTS = or(DVTSo1,DVTSo2)
        )
    ).
rtype_up_1([],LHS,RHS,DVTS) :-
    type_UpApprox(RHS,[],Type),
    (  Type = top -> DVTS=[] 
     ; etype_tabled(Type) -> DVTS=vtbot
     ; DVTS=[LHS/Type]
    ).
rtype_up_1([H|VT],LHS,RHS,DVTS) :-
    type_UpApprox(RHS,[H|VT],Type),
    (  Type = top -> DVTS=[H|VT] 
     ; vt_strengthen_v2([H|VT],LHS,Type,DVTS) -> true
     ; DVTS=vtbot
    ).



/* rtype_bipai(+Call,+ASubi,-ASubo) ASubo describes the set of
   substitutions that satisfy ASubi and that are answers to Call.
*/

rtype_bipai(copy,ASubi,ASubi). 
rtype_bipai(solve(E0),ASubi,ASubo) :- 	rtype_solve(E0,ASubi,ASubo).
rtype_bipai(glb(VTS),ASubi,ASubo)  :-   vtsglb(VTS,ASubi,ASubo). 
rtype_bipai(glb_solve(VTS,E0),ASubi,ASubo)  :-   
           vtsglb(VTS,ASubi,ASub1),rtype_solve(E0,ASub1,ASubo). 
rtype_bipai(var(V),ASubi,ASubo)    :- 
           rtype_bipai_var(ASubi,V,ASubo).
           % keep those variable typings in which V is of type "top" 

    rtype_bipai_var(vtbot,_,vtbot).
    rtype_bipai_var(or(DVTS1,DVTS2),V,DVTS) :-
        rtype_bipai_var(DVTS1,V,DVTSo1),
        (
            DVTSo1 == vtbot
        -> 
            rtype_bipai_var(DVTS2,V,DVTS)
        ;
            rtype_bipai_var(DVTS2,V,DVTSo2),
            (
                 DVTSo2 == vtbot 
            ->
                 DVTS = DVTSo1
            ;
                 DVTS = or(DVTSo1,DVTSo2)
            )
        ).
    rtype_bipai_var([],_,[]).
    rtype_bipai_var([H|VT],V,DVTS) :- 
        type_UpApprox(V,[H|VT],Tp),
        (
            etype_tabled(not(Tp)) % Tp is equivalent to top ?
        ->
            DVTS=vtbot
	 ; 
            DVTS=[H|VT]
        ). 

/* rtype_compile(+Call,-Compiled) compiles a bip call to its compiled
   form to factor out computation that is independent of abstract
   input state. */

rtype_compile(abolish(_),       copy).
rtype_compile(assert(_),        copy).
rtype_compile(retract(_),       copy).
rtype_compile(erase(I), glb(VTS)) :- 
    vts_up_v2(integer,I,[],VTS).
rtype_compile(recorda(_Key,_Term,Ref), glb(VTS))    :-
    vts_up_v2(integer,Ref,[],VTS).
rtype_compile(recorded(_Key,_Term,Ref), glb(VTS))    :-
    vts_up_v2(integer,Ref,[],VTS).
rtype_compile(format(Fmt), glb(VTS))      :-
    vts_up_v2(or(atom, list(integer)),Fmt,[],VTS).
rtype_compile(format(Fmt,_), glb(VTS))    :-
    vts_up_v2(or(atom, list(integer)),Fmt,[],VTS).
rtype_compile(format(_,Fmt,_), glb(VTS))  :-
    vts_up_v2(or(atom, list(integer)),Fmt,[],VTS).


rtype_compile(repeat,           copy).
rtype_compile(listing,          copy).
rtype_compile(listing(_),       copy).
rtype_compile(compound(_),      copy).
rtype_compile(portray_clause(_),copy).
rtype_compile(ground(_),        copy).
rtype_compile(!,	        copy).
rtype_compile(display(_),	copy).
rtype_compile(put(_),           copy).
rtype_compile(nl,               copy).
rtype_compile(print(_),         copy).
rtype_compile(write(_),         copy).
rtype_compile(writeq(_),	copy).
rtype_compile(keysort(_,_),     copy). % sorting on term order
rtype_compile(sort(_,_),	copy).
rtype_compile(@<(_,_),	        copy). % comparing on term order
rtype_compile(@>(_,_),	        copy).
rtype_compile(@=<(_,_),	        copy).
rtype_compile(@>=(_,_),         copy).
rtype_compile(true,             copy).
rtype_compile(nonvar(_),	copy).
rtype_compile(read(_),	        copy).
rtype_compile(\==(_,_),         copy).
rtype_compile(\=(_,_),          copy).
rtype_compile(=\=(_,_),	        copy).
rtype_compile(statistics(_N1,_N2), copy).
rtype_compile(compare(_,_,_), copy).


% these built-ins change type information as fail

%rtype_compile(abort,            _) :- fail.
%rtype_compile(fail,             _) :- fail.
%rtype_compile(false,	         _) :- fail.

% these built-ins change type information as unification

rtype_compile(=(Left,Right),    solve(E0))     :-
	mgu(Left,Right,E0).
rtype_compile(=@=(Left,Right),  solve(E0))     :-
	mgu(Left,Right,E0).
        % structurally identical terms must unify
rtype_compile(==(Left,Right),   solve(E0))     :-
	mgu(Left,Right,E0).
	% identical terms must unify
rtype_compile(is(Left,Right),   solve(E0))     :-
	mgu(Left,Right,E0).
	% identical terms must unify

% these built-ins type check their arguments

rtype_compile(float(Tm), glb(VTS))     :-
        vts_up_v2(float,Tm,[],VTS).
rtype_compile(string(Tm), glb(VTS))    :-
        vts_up_v2(string,Tm,[],VTS).
rtype_compile(atom(Tm),  glb(VTS))     :-
        vts_up_v2(atom,Tm,[],VTS).
rtype_compile(atomic(Tm), glb(VTS))    :-
        vts_up_v2(atomic,Tm,[],VTS).
rtype_compile(integer(Tm), glb(VTS))   :-
        vts_up_v2(integer,Tm,[],VTS).
rtype_compile(number(Tm), glb(VTS))    :-
        vts_up_v2(number,Tm,[],VTS).
rtype_compile(tab(Tm), glb(VTS))       :-
        vts_up_v2(integer,Tm,[],VTS).
rtype_compile(<(N1,N2), glb(VTS))      :-
        vts_up_v2(number,N1,[],VTS1),
        vts_up_v2(number,N2,VTS1,VTS).
rtype_compile(>(N1,N2), glb(VTS))     :-
        vts_up_v2(number,N1,[],VTS1),
        vts_up_v2(number,N2,VTS1,VTS).
rtype_compile(=<(N1,N2), glb(VTS))     :-
        vts_up_v2(number,N1,[],VTS1),
        vts_up_v2(number,N2,VTS1,VTS).
rtype_compile(>=(N1,N2), glb(VTS))     :-
        vts_up_v2(number,N1,[],VTS1),
        vts_up_v2(number,N2,VTS1,VTS).
rtype_compile(length(L,N),  glb(VTS))     :-
        vts_up_v2(integer,N,[],VTS1),
        vts_up_v2(list(top),L,VTS1,VTS).
rtype_compile(=..(_,L), glb(VTS)) :-
        vts_up_v2(list(top),L,[],VTS).
rtype_compile(functor(_,Nm,Ar),  glb(VTS))     :-
        vts_up_v2(integer,Ar,[],VTS1),
        vts_up_v2(atom,Nm,VTS1,VTS).
rtype_compile(arg(N,_,_), glb(VTS))     :- % may be made more precise??
        vts_up_v2(integer,N,[],VTS).
rtype_compile(name(AI,S), glb(VTS))     :-
        vts_up_v2(or(atom,integer),AI,[],VTS1),
        vts_up_v2(list(integer),S,VTS1,VTS).
rtype_compile(=:=(N1,N2), glb(VTS))     :- 
        vts_up_v2(number,N1,[],VTS1),
        vts_up_v2(number,N2,VTS1,VTS).
rtype_compile(var(V), var(V)) :- 
        !, 
        variable(V).


/* rtype_bip_has_cond(+Call,-Cond) is true iff +Call has type
   condition -Cond.  */

%rtype_bip_has_cond(assert(_),        _) :- fail.
%rtype_bip_has_cond(retract(_),       _) :- fail.
%rtype_bip_has_cond(repeat,           _) :- fail.
%rtype_bip_has_cond(listing,          _) :- fail.
%rtype_bip_has_cond(listing(_),       _) :- fail.
%rtype_bip_has_cond(compound(_),      _) :- fail.
%rtype_bip_has_cond(ground(_),        _) :- fail.
%rtype_bip_has_cond(!,	        _) :- fail.
%rtype_bip_has_cond(display(_),	_) :- fail.
%rtype_bip_has_cond(nl,               _) :- fail.
%rtype_bip_has_cond(print(_),         _) :- fail.
%rtype_bip_has_cond(write(_),         _) :- fail.
%rtype_bip_has_cond(writeq(_),	_) :- fail.
%rtype_bip_has_cond(abort,            _) :- fail.
%rtype_bip_has_cond(fail,             _) :- fail.
%rtype_bip_has_cond(false,            _) :- fail.
%rtype_bip_has_cond(@<(_,_),	      _) :- fail. 
%rtype_bip_has_cond(@>(_,_),	        _) :- fail.
%rtype_bip_has_cond(@=<(_,_),	        _) :- fail.
%rtype_bip_has_cond(@>=(_,_),         _) :- fail.
%rtype_bip_has_cond(true,             _) :- fail.
%rtype_bip_has_cond(nonvar(_),	_) :- fail.
%rtype_bip_has_cond(read(_),	_) :- fail.
%rtype_bip_has_cond(\==(_,_),   _) :- fail.
%rtype_bip_has_cond(\=(_,_),    _) :- fail.
%rtype_bip_has_cond(=\=(_,_),	_) :- fail.
%rtype_bip_has_cond(var(_), _)     :- fail.
%rtype_bip_has_cond(=(_,_), _)     :- fail.
%rtype_bip_has_cond(=@=(_,_), _)   :- fail.
%rtype_bip_has_cond(==(_,_),_)     :- fail.
%rtype_bip_has_cond(is(_,_), _)    :- fail.
%rtype_bip_has_cond(float(_), _)   :- fail.
%rtype_bip_has_cond(string(_),_)   :- fail.
%rtype_bip_has_cond(atom(_),_)     :- fail.
%rtype_bip_has_cond(atomic(_),_)   :- fail.
%rtype_bip_has_cond(integer(_),_)  :- fail.
%rtype_bip_has_cond(number(_),_)   :- fail.
%rtype_bip_has_cond(length(_,_),_)  :- fail.
%rtype_bip_has_cond(=..(_,_),_)     :- fail.
%rtype_bip_has_cond(portray_clause(_),_) :- fail.
%rtype_bip_has_cond(compare(_,_,_), _)   :- fail.


rtype_bip_has_cond(erase(I),VTS)  :- 
    vts_LowApprox(integer,I,VTS).
rtype_bip_has_cond(format(Fmt), VTS)      :-
    vts_LowApprox(or(atom, or(string,list(integer))),Fmt,VTS).
rtype_bip_has_cond(format(Fmt,_), VTS)    :-
    vts_LowApprox(or(atom, or(string,list(integer))),Fmt,VTS).
rtype_bip_has_cond(format(_,Fmt,_), VTS)  :-
    vts_LowApprox(or(atom, or(string,list(integer))),Fmt,VTS).
rtype_bip_has_cond(keysort(Li,_), VTS) :- 
    vts_LowApprox(list(pair(top,top)),Li,VTS).
rtype_bip_has_cond(sort(Li,_), VTS)    :- 
    vts_LowApprox(list(top),Li,VTS).

% following conditions can be relaxed in a more powerful type system.

rtype_bip_has_cond(put(N),VTS) :- vts_LowApprox(integer,N,VTS).
rtype_bip_has_cond(statistics(Key,_), VTS) :- 
    vts_LowApprox(atom,Key,VTS). % needs be refined

rtype_bip_has_cond(functor(_,_,_),_) :-
    write("The program contains functor/3!"),nl,halt.
rtype_bip_has_cond(arg(_,_,_),_) :- 
    write("The program contains arg/3!"),nl,halt.
rtype_bip_has_cond(abolish(_),_) :- 
     write("The program contains abolish/1!"),nl,halt.
rtype_bip_has_cond(recorda(_,_,_),_)    :-
     write("The program contains recorda/3!"),nl,halt.
rtype_bip_has_cond(recorded(_,_,_),_)   :-
    write("The program contains recorded/3!"),nl,halt.

% these built-ins type check their arguments

rtype_bip_has_cond(tab(Tm), VTS)       :-
        vts_LowApprox(integer,Tm,VTS).
rtype_bip_has_cond(<(N1,N2), VTS)      :-
        vts_LowApprox(number,N1,VTS1),
        vts_LowApprox(number,N2,VTS2),
        vtsglb(VTS1,VTS2,VTS). 
rtype_bip_has_cond(>(N1,N2), VTS)      :-
        vts_LowApprox(number,N1,VTS1),
        vts_LowApprox(number,N2,VTS2),
        vtsglb(VTS1,VTS2,VTS). 
rtype_bip_has_cond(=<(N1,N2), VTS)     :-
        vts_LowApprox(number,N1,VTS1),
        vts_LowApprox(number,N2,VTS2),
        vtsglb(VTS1,VTS2,VTS). 
rtype_bip_has_cond(>=(N1,N2), VTS)     :-
        vts_LowApprox(number,N1,VTS1),
        vts_LowApprox(number,N2,VTS2),
        vtsglb(VTS1,VTS2,VTS). 
rtype_bip_has_cond(=:=(N1,N2),VTS)     :- 
        vts_LowApprox(number,N1,VTS1),
        vts_LowApprox(number,N2,VTS2),
        vtsglb(VTS1,VTS2,VTS).
rtype_bip_has_cond(name(AI,S), VTS)    :-
        vts_LowApprox(or(atom,integer),AI,VTS1),
        vts_LowApprox(string,S,VTS2),
        append(VTS1,VTS2,VTS). 




/* rtype_leq(+VTS1,+VTS2) VTS1 is subsumed by VTS2 */

rtype_leq(vtbot,_).
rtype_leq(or(DVTS11,DVTS12),DVTS2) :- 
    rtype_leq(DVTS11,DVTS2),
    rtype_leq(DVTS12,DVTS2).
rtype_leq([],DVTS) :- % [] is the top variable typing
    dvts_subsumes_vt(DVTS,[]).
rtype_leq([H|L],DVTS) :- 
    dvts_subsumes_vt(DVTS,[H|L]). 

    /* dvts_subsumes_vt(+DVTS,+VT) VT is subsumed by DVTS */

    dvts_subsumes_vt(vtbot,VT) :- % VT is less than vtbot
        vt_bot(VT).
    dvts_subsumes_vt(or(DVTS1,DVTS2),VT) :- 
        ( dvts_subsumes_vt(DVTS1,VT) -> true ; dvts_subsumes_vt(DVTS2,VT) ).
    dvts_subsumes_vt([],_VT). % [] is the top variable typing
    dvts_subsumes_vt([H|L],VT) :- 
        vt_leq(VT,[H|L]).

    vt_bot([_/T|VT]) :- 
        ( etype_tabled(T) -> true; vt_bot(VT) ).


vt_leq([],[])    :- !.
vt_leq([],[H|L]) :- vt_id([H|L]).  %%%
vt_leq([_|_],[]) :- !.
vt_leq([P1/T1|VT1],[P2/T2|VT2]) :-
	(
		P1 @< P2 
	->      % skip P1 as VT2(P1) = top
		vt_leq(VT1,[P2/T2|VT2])
	;
		P1 @> P2
	->      % 
                etype_tabled(not(T2)), % T2 must be equivalent to "top"
		vt_leq([P1/T1|VT1],VT2)
	;
                % test T1<=T2
                etype_tabled(and(T1,not(T2))),
		vt_leq(VT1,VT2)
	).

vt_id([]).
vt_id([_/T|VT]) :- 
    etype_tabled(not(T)),
    vt_id(VT). 



%% Following code is for dealing with negative types 

/* negate_atom_UpApprox(A,R) R is an upper approximation to not(A). R is a
   disjunction of atomic types. */

negate_atom_UpApprox(Atom,R) :- 
    typedef_constructors(Cnsts),
    functor(Atom,Cnst,Arity),
    delete(Cnsts,Cnst/Arity,Cnsts0), delete(Cnsts0,top/0,Cnsts1),
    most_gen_types(Cnsts1,bot,R1),
    Atom =.. [Cnst|Seq],
    ( 
      Seq = [] 
    -> 
      R = R1 
     ;
      negate_seq(Seq,DSeq),
      apply_cnst_to_dseq(DSeq,Cnst,R2),
      tor(R1,R2,R)
    ).

    most_gen_types([],R,R).
    most_gen_types([Cnst/Arity|Cnsts],Racc,R) :-
        tops(Arity,Ones),
        A =..[Cnst|Ones],
        tor(A,Racc,Racc1),
        most_gen_types(Cnsts,Racc1,R).


        tops(0,[]).
        tops(N,[top|Tops]) :-
            N > 0,
            N1 is N-1,
            tops(N1,Tops).

    apply_cnst_to_dseq(or(DSeq1,DSeq2),Cnst,R) :- 
        !,
        apply_cnst_to_dseq(DSeq1,Cnst,R1),
        apply_cnst_to_dseq(DSeq2,Cnst,R2),
        tor(R1,R2,R).
    apply_cnst_to_dseq(Seq,Cnst,R) :- 
        R =.. [Cnst|Seq]. 

/* negate_atom_LowApprox(A,R) R is an Lower approximation to not(A). R
   is a disjunction of atomic types. */

negate_atom_LowApprox(Atom,R) :- 
    typedef_constructors(Cnsts),
    functor(Atom,Cnst,Arity),
    ( setof(NA,overlap_cnst(Cnst/Arity,NA),Overlaps) -> true ; Overlaps =[] ),
    delete_all([Cnst/Arity,top/0|Overlaps],Cnsts,Cnsts1),
    most_gen_types(Cnsts1,bot,R1),
    Atom =.. [Cnst|Seq],
    ( 
      Seq = [] 
    -> 
      R = R1 
     ;
      negate_seq(Seq,DSeq),
      apply_cnst_to_dseq(DSeq,Cnst,R2),
      tor(R1,R2,R)
    ).

    overlap_cnst(Cnst1/Arity1,Cnst2/Arity2) :- 
        functor(AT1,Cnst1,Arity1),
        typedef(Fn,_,AT1),
        typedef(Fn,_,AT2),
        functor(AT2,Cnst2,Arity2).

    delete_all([],L,L).
    delete_all([D|Ds],Li,Lo) :- 
        delete(Li,D,Li1),
        delete_all(Ds,Li1,Lo).

bigtglb([],top).
bigtglb([Tp],Tp).
bigtglb([Tp1,Tp2|Tps],Tp) :- 
   bigtglb([Tp2|Tps],Tp1,Tp).


bigtglb([],Tp,Tp).
bigtglb([A|As],Tp0,Tp) :-
    tand(A,Tp0,Tp1),
    bigtglb(As,Tp1,Tp).

bigtlub(Tps,Tp) :- 
   bigtlub(Tps,bot,Tp).

bigtlub([],Tp,Tp).
bigtlub([A|As],Tp0,Tp) :-
    tor(A,Tp0,Tp1),
    bigtlub(As,Tp1,Tp).






/* mapsto(+Key,+Map,-Value) If +Key is in the domain of +Map then
   mapsto(Key,Map,Value) succeeds and Value is the value assigned to
   Key by Map. Otherwise mapsto(Key,Map,Value) fails. */

mapsto(Key,[Key1/Value1|Map],Value) :-
	( Key==Key1 -> Value=Value1 ; mapsto(Key,Map,Value) ).

prim_or_negated(atom).
prim_or_negated(atomic).
prim_or_negated(integer).
prim_or_negated(float).
prim_or_negated(number).
prim_or_negated(not(X)) :- prim_type(X).

tand(T1,T2,T3)  :- 
(    ( prim_or_negated(T1), prim_or_negated(T2) ) 
->
    tand_prim(T1,T2,T3)
;
    ( 
        etype_tabled(and(T1,not(T2))) % T1<=T2
    ->  
        T3 = T1
    ; 
        etype_tabled(and(T2,not(T1))) % T2<=T1
    -> 
        T3 = T2
    ;
        etype_tabled(and(T1,T2)) 
     -> 
        T3 = bot
    ;   
        T3 = and(T1,T2) 
    )
).

tor(T1,T2,T3)  :- 
    ( prim_or_negated(T1), prim_or_negated(T2) ) 
->
    tor_prim(T1,T2,T3)
;
    ( 
        etype_tabled(and(T1,not(T2))) % T1<=T2
    ->  
        T3 = T2
    ; 
        etype_tabled(and(T2,not(T1))) % T2<=T1
    -> 
        T3 = T1
    ;
        T3 = or(T1,T2) 
    ).

tleq(T1,T2) :- etype_tabled(and(T1,not(T2))).

% tabling etype, using XSB, we can expect much better performance


:- dynamic '$etype_tabled'/2. 

rtype_init :- retractall('$etype_tabled'(_,_)).


/* Use this codes for tests with tabling */

etype_tabled(T) :- 
    inc_counter('total_emptiness_checks'),
    (
        '$etype_tabled'(T,Ans) 
    -> 
        Ans = yes
    ;
        inc_counter('distinct_emptiness_checks'),
        current_time(TStart),
        ( 
            etype(T) 
        -> 
            current_time(TEnd),
            Lapse is TEnd-TStart,
            add_counter(checking_time,Lapse),
            assert('$etype_tabled'(T,yes)) 
	 ;
            current_time(TEnd),
            Lapse is TEnd-TStart,
            add_counter(checking_time,Lapse),
            assert('$etype_tabled'(T,no)),
            !,
            fail 
        )
    ).

/*
% This code doesn't memoize. It just measures time 
 
etype_tabled(T) :- 
    inc_counter('total_emptiness_checks'),
    (
        current_time(TStart),
        ( 
            etype(T) 
        -> 
            current_time(TEnd),
            Lapse is TEnd-TStart,
            add_counter(checking_time,Lapse)
	 ;
            current_time(TEnd),
            Lapse is TEnd-TStart,
            add_counter(checking_time,Lapse),
            !,
            fail 
        )
    ).

*/


make_map([],[],[]).
make_map([Key|Keys],[Val|Vals],[Key/Val|Map]) :- 
    make_map(Keys,Vals,Map).

remove_vt_bot(vtbot,vtbot).
remove_vt_bot(or(DVTS11,DVTS12),DVTS) :- 
    remove_vt_bot(DVTS11,DVTSo1),
    (
        DVTSo1 == vtbot
    -> 
        remove_vt_bot(DVTS12,DVTS)
    ;
        remove_vt_bot(DVTS12,DVTSo2),
        (
             DVTSo2 == vtbot 
        ->
             DVTS = DVTSo1
        ;
             DVTS = or(DVTSo1,DVTSo2)
        )
    ).
remove_vt_bot([],[]). % [] is the top variable typing
remove_vt_bot([H|L],DVTSo) :-
    ( vt_bot([H|L])-> DVTSo = vtbot ; DVTSo=[H|L] ). 


remove_vts_subsumed_by(vtbot,_,vtbot). 
remove_vts_subsumed_by(or(DVTS11,DVTS12),DVTS2,DVTS) :- 
    remove_vts_subsumed_by(DVTS11,DVTS2,DVTSo1),
    (
        DVTSo1 == vtbot
    -> 
        remove_vts_subsumed_by(DVTS12,DVTS2,DVTS)
    ;
        remove_vts_subsumed_by(DVTS12,DVTS2,DVTSo2),
        (
             DVTSo2 == vtbot 
        ->
             DVTS = DVTSo1
        ;
             DVTS = or(DVTSo1,DVTSo2)
        )
    ).
remove_vts_subsumed_by([],DVTS2,DVTS) :- 
    dvts_subsumes_vt(DVTS2,[]) -> DVTS = vtbot ; DVTS=[].
remove_vts_subsumed_by([H|L],DVTS2,DVTS) :- 
    dvts_subsumes_vt(DVTS2,[H|L]) -> DVTS = vtbot ; DVTS=[H|L].

remove_subsumed_vt(vtbot,vtbot).
remove_subsumed_vt([],[]).
remove_subsumed_vt([H|L],[H|L]).
remove_subsumed_vt(or(DVTS1,DVTS2),DVTS) :- 
    remove_subsumed_vt(DVTS1,DVTSo1),
    remove_subsumed_vt(DVTS2,DVTSo2),
    remove_vts_subsumed_by(DVTSo1,DVTSo2,DVTSo3),
    (
        DVTSo3 == vtbot 
    ->
        DVTS = DVTSo2
     ; 
        remove_vts_subsumed_by(DVTSo2,DVTSo3,DVTSo4),
        (
            DVTSo4 == vtbot
        ->
            DVTS = DVTSo3
        ;
            DVTS = or(DVTSo3,DVTSo4)
        )
    ).


/* rtype_join(+VTS1,+VTS2,-VTS) VTS=VTS1+VTS2 redundency removed.
    This implementation makes use the fact that VTS2 is already
    redundency-free and VTS1 contains no variable typing that denotes
    the empty set of substitutions.  */

rtype_join(VTS1, VTS2, VTSo) :-
    remove_subsumed_vt(VTS1,VTSa),
    remove_vts_subsumed_by(VTSa,VTS2,VTSo1),
    (
        VTSo1 == vtbot
    -> 
        VTSo = VTS2
    ;
        remove_vts_subsumed_by(VTS2,VTSo1,VTSo2),
        ( VTSo2==vtbot -> VTSo=VTSo1 ; VTSo=or(VTSo2,VTSo1) )
    ).

reduce(vtbot,vtbot). 
reduce([],[]).
reduce([H|L],DVTS) :- 
    ( vt_bot([H|L]) -> DVTS = vtbot ; DVTS=[H|L] ). 
    % check for empty variable typing
reduce(or(DVTS1,DVTS2),DVTS) :- 
    reduce(DVTS1,DVTSo1),
    reduce(DVTS2,DVTSo2),
    remove_vts_subsumed_by(DVTSo1,DVTSo2,DVTSo3),
    (
        DVTSo3 == vtbot 
    ->
        DVTS = DVTSo2
     ; 
        remove_vts_subsumed_by(DVTSo2,DVTSo3,DVTSo4),
        (
            DVTSo4 == vtbot
        ->
            DVTS = DVTSo3
        ;
            DVTS = or(DVTSo3,DVTSo4)
        )
    ).

/* vts_up_v2(+Rt,+Tm,+VTSi,-VTSo) propagates the type Rt of Tm
   downwards the term structure and obtains a disjunction of variable
   typing and uses it to strength each of those in VTSi to give
   VTSo. */

vts_up_v2(and(R1,R2), Tm, VTSi, VTSo) :- 
%%always called with first argument begin in disjuctive normal form
    !,
    vts_up_v2(R1, Tm, VTSi, VTS1),
    vts_up_v2(R2, Tm, VTS1, VTSo).

vts_up_v2(or(R1,R2),Tm, VTSi, or(VTS1,VTS2)) :- 
    !,
    vts_up_v2(R1, Tm, VTSi, VTS1),
    vts_up_v2(R2, Tm, VTSi, VTS2).

vts_up_v2(top, _, VTS, VTS) :- 
    !.

vts_up_v2(R,V,VTSi,VTSo) :-
    variable(V),
    !,
    vts_strengthen(VTSi,V,R,VTSo).

vts_up_v2(not(A), Tm, VTSi, VTSo) :- 
    !,
    negate_atom_UpApprox(A,R),
    vts_up_v2(R, Tm, VTSi, VTSo).

vts_up_v2(ATp, Tm, VTSi, VTSo) :- 
    functor(Tm,Name,Arity),
    ( 
        chk_membership(Name,Arity,ATp)
    -> 
        VTSo = VTSi
    ;
        Tm =.. [_|Args],
        findall(TArgs, typedef(Name/Arity, TArgs, ATp), TArgsSet),
        vts_up_v2_rules(TArgsSet,Args,VTSi,VTSo)
    ).

    chk_membership(Name,0,ATp) :- 
        (   ATp = atomic -> atomic(Name) 
          ; ATp = atom -> atom(Name) 
          ; ATp = number -> number(Name) 
          ; ATp = float -> float(Name) 
          ; ATp = integer -> integer(Name) 
        ).

   vts_up_v2_rules([],_Args,_VTSi,vtbot). % no rule
   vts_up_v2_rules([TArgs|TArgsSet],Args,VTSi,VTSo) :- 
       vts_up_v2_one_rule_args(TArgs,Args,VTSi,VTSacc),
       vts_up_v2_rules(TArgsSet,Args,VTSi,VTSacc,VTSo).
   

   vts_up_v2_rules([],_Args,_VTSi,VTS,VTS).
   vts_up_v2_rules([TArgs|TArgsSet],Args,VTSi,VTSacc,VTSo) :- 
       vts_up_v2_one_rule_args(TArgs,Args,VTSi,VTS1),
       vts_up_v2_rules(TArgsSet,Args,VTSi,or(VTS1,VTSacc),VTSo).

        vts_up_v2_one_rule_args([],[],VTS, VTS). 
        vts_up_v2_one_rule_args([R|Rs],[Tm|Tms],VTSi,VTS) :-
            type_dnf(R,DC), 
            % make sure vts_up_v2 is called with first argument in DNF
            vts_up_v2(DC,Tm, VTSi, VTS1),
            vts_up_v2_one_rule_args(Rs,Tms,VTS1,VTS).

vts_strengthen(vtbot,_,_,vtbot).
vts_strengthen([],V,R,[V/R]).
vts_strengthen([H|L],V,R,DVTS) :- 
    ( vt_strengthen_v2([H|L],V,R,DVTS) -> true ; DVTS=vtbot ).
vts_strengthen(or(DVTS11,DVTS12),V,R,DVTS) :- 
    vts_strengthen(DVTS11,V,R,DVTSo1),
    (
        DVTSo1 == vtbot
    -> 
        vts_strengthen(DVTS12,V,R,DVTS)
    ;
        vts_strengthen(DVTS12,V,R,DVTSo2),
        (
             DVTSo2 == vtbot 
        ->
             DVTS = DVTSo1
        ;
             DVTS = or(DVTSo1,DVTSo2)
        )
    ).

/* vt_strengthen_v2(+VTi,+V,+T,-VTo) VTo = VT1[Var/(VTi(V) and T)] and
   the VTo does not denote the empty set of substitution. */

vt_strengthen_v2([],V,T,[V/T]).
vt_strengthen_v2([Key1/Type1|VT1],Key2,Type2,[Key3/Type3|VT3]) :-
	(
		Key1 @< Key2
	->
		Key3 = Key1,Type3=Type1,
		vt_strengthen_v2(VT1,Key2,Type2,VT3)
	;
		Key1 @> Key2
	->
		\+ etype_tabled(Type2),
                Key3 = Key2,Type3=Type2,VT3=[Key1/Type1|VT1]		
	;
		tand(Type1,Type2,Type3), 
	        \+ etype_tabled(Type3),
                Key3 = Key2, VT3=VT1
	).


vts_LowApprox(and(R1,R2),Tm,VTS) :- 
%%always called with first argument begin in disjuctive normal form
%%only the clause for negative types differs from that for upper approx 
    !,
    vts_LowApprox(R1,Tm,VTS1),
    vts_LowApprox(R2,Tm,VTS2),
    vtsglb(VTS1,VTS2,VTS).

vts_LowApprox(or(R1,R2),Tm,or(VTS1,VTS2)) :- 
    !,
    vts_LowApprox(R1,Tm,VTS1),
    vts_LowApprox(R2,Tm,VTS2).

vts_LowApprox(top,_,[]) :- 
    !.

vts_LowApprox(R,V,[V/R]) :-
    variable(V),
    !.

vts_LowApprox(not(A),Tm,VTS) :- 
    !,
    negate_atom_LowApprox(A,R),
    vts_LowApprox(R,Tm,VTS).


vts_LowApprox(ATp,Tm,VTS) :- 
    functor(Tm,Name,Arity),
    Tm =.. [_|TArgs],
    findall(VTSr,vts_LowApprox_one_rule(Name,Arity,TArgs,ATp,VTSr), VTSrSet),
    join_all(VTSrSet,VTS).

    join_all([],vtbot).
    join_all([One],One).
    join_all([One,Two|Others],or(One,DVTS)) :- join_all([Two|Others],DVTS).

    vts_LowApprox_one_rule(Name,0,[],ATp,[]) :- % non-deterministic 
        prim_type(ATp),     % extra clause for contants and primitive types
        membership_prim(ATp,Name),
        !.            % do not care about other variable typings

    vts_LowApprox_one_rule(Name,Arity,Targs,ATp,VTSr) :- % non-deterministic
        typedef(Name/Arity, Tis, ATp),               % multiple type rules
        vts_LowApprox_one_rule_args(Tis,Targs,[], VTSr). 
                                             % [] is top variable typing
                                         
        vts_LowApprox_one_rule_args([],[],VTSr, VTSr). 
        vts_LowApprox_one_rule_args([R|Rs],[Tm|Tms],VTSr0, VTSr) :-
            type_dnf(R,DC), 
            % make sure vts_LowApprox is called with first argument in DNF
            vts_LowApprox(DC,Tm,VTSr1),
            vtsglb(VTSr0,VTSr1,VTSr2),
            vts_LowApprox_one_rule_args(Rs,Tms,VTSr2,VTSr),
            !.



/* rtype_bw_compile(+Call,-Compiled) compiles a bip call to its
   compiled form to factor out computation that is independent of
   abstract input state. Success of a builtin call is a conjunction
   (Sd and S~d) where Sd is implied by the demand of the builtin
   call. Compiled is set to S~d for most builtins. */

rtype_bw_compile(recorda(_Key,_Term,Ref), complement(VTS))    :-
    vts_up_v2(integer,Ref,[],VTS).
rtype_bw_compile(recorded(_Key,_Term,Ref), complement(VTS))    :-
    vts_up_v2(integer,Ref,[],VTS).

rtype_bw_compile(abolish(_),       copy).
rtype_bw_compile(assert(_),        copy).
rtype_bw_compile(retract(_),       copy).
rtype_bw_compile(erase(_I),        copy). 
rtype_bw_compile(format(_Fmt),     copy).
rtype_bw_compile(format(_Fmt,_),   copy).
rtype_bw_compile(format(_,_Fmt,_), copy).
rtype_bw_compile(put(_N),          copy).

rtype_bw_compile(keysort(_L,R), complement(VTS)) :- 
    vts_up_v2(list(pair(top,top)), R,[],VTS).
rtype_bw_compile(sort(_L,R), complement(VTS)) :- 
    vts_up_v2(list(top), R,[],VTS).
rtype_bw_compile(statistics(_N1,N2), complement(VTS)) :- 
     vts_up_v2(list(integer),N2,[],VTS). 
rtype_bw_compile(compare(A,_,_), VTS) :- 
     vts_up_v2(atom, A, [],VTS).


rtype_bw_compile(repeat,           copy).
rtype_bw_compile(listing,          copy).
rtype_bw_compile(listing(_),       copy).
rtype_bw_compile(compound(_),      copy).
rtype_bw_compile(portray_clause(_),copy).
rtype_bw_compile(ground(_),        copy).
rtype_bw_compile(!,	        copy).
rtype_bw_compile(display(_),	copy).
rtype_bw_compile(nl,            copy).
rtype_bw_compile(print(_),      copy).
rtype_bw_compile(write(_),      copy).
rtype_bw_compile(writeq(_),	copy).
rtype_bw_compile(@<(_,_),	copy). % comparing on term order
rtype_bw_compile(@>(_,_),	copy).
rtype_bw_compile(@=<(_,_),	copy).
rtype_bw_compile(@>=(_,_),      copy).
rtype_bw_compile(true,          copy).
rtype_bw_compile(nonvar(_),	copy).
rtype_bw_compile(read(_),	        copy).
rtype_bw_compile(\==(_,_),         copy).
rtype_bw_compile(\=(_,_),          copy).
rtype_bw_compile(=\=(_,_),	        copy).
rtype_bw_compile(var(_V), copy).


% these built-ins change type information as fail

%rtype_bw_compile(abort,            _) :- fail.
%rtype_bw_compile(fail,             _) :- fail.
%rtype_bw_compile(false,	         _) :- fail.

% these built-ins change type information as unification

rtype_bw_compile(=(Left,Right),    dsolve(E0))     :-
	mgu(Left,Right,E0).
rtype_bw_compile(=@=(Left,Right),  dsolve(E0))     :-
	mgu(Left,Right,E0).
        % structurally identical terms must unify
rtype_bw_compile(==(Left,Right),   dsolve(E0))     :-
	mgu(Left,Right,E0).
	% identical terms must unify
rtype_bw_compile(is(Left,Right),   dsolve(E0))     :-
	mgu(Left,Right,E0).
	% identical terms must unify

% these built-ins type check their arguments

rtype_bw_compile(float(Tm), complement(VTS))     :-
        vts_up_v2(float,Tm,[],VTS).
rtype_bw_compile(atom(Tm),  complement(VTS))     :-
        vts_up_v2(atom,Tm,[],VTS).
rtype_bw_compile(atomic(Tm), complement(VTS))    :-
        vts_up_v2(atomic,Tm,[],VTS).
rtype_bw_compile(integer(Tm), complement(VTS))   :-
        vts_up_v2(integer,Tm,[],VTS).
rtype_bw_compile(number(Tm), complement(VTS))    :-
        vts_up_v2(number,Tm,[],VTS).
rtype_bw_compile(tab(Tm), complement(VTS))       :-
        vts_up_v2(integer,Tm,[],VTS).

rtype_bw_compile(<(_,_), copy).
rtype_bw_compile(>(_,_), copy).
rtype_bw_compile(=<(_,_),copy).
rtype_bw_compile(>=(_,_),copy).

rtype_bw_compile(length(L,N),  complement(VTS))     :-
        vts_up_v2(integer,N,[],VTS1),
        vts_up_v2(list(top),L,VTS1,VTS).
rtype_bw_compile(=..(_,L), complement(VTS)) :-
        vts_up_v2(list(top),L,[],VTS).
rtype_bw_compile(functor(_,Nm,Ar),  complement(VTS))     :-
        vts_up_v2(integer,Ar,[],VTS1),
        vts_up_v2(atom,Nm,VTS1,VTS).
rtype_bw_compile(arg(N,_,_), complement(VTS))     :- 
        vts_up_v2(integer,N,[],VTS).
rtype_bw_compile(name(AI,S), complement(VTS))     :-
        vts_up_v2(or(atom,integer),AI,[],VTS1),
        vts_up_v2(string,S,VTS1,VTS).
rtype_bw_compile(=:=(_,_), copy).



/* rtype_bw_bipai(+Call,+ASubi,-ASubo) ASubo describes the set of
   substitutions that satisfy ASubi and that are answers to Call.
*/

rtype_bw_bipai(copy,ASubi,ASubi). 
rtype_bw_bipai(complement(VTS),ASubi,ASubo)  :- 
    pseduo_complement(VTS,ASubi,ASubo). 
rtype_bw_bipai(dsolve(E0),ASubi,ASubo) :- 
    rtype_solve_bw(E0,ASubi,ASubo).




/* vts_negate(VTSi,VTSo) VTSo = ~VTSi */

vts_negate(vtbot,[]).
vts_negate([],vtbot). 
vts_negate([X/T],[X/NotT]) :- 
    type_dnf(not(T), NotT).
vts_negate([X/T,XT|VT],or([X/NotT],VTS)) :- 
    type_dnf(not(T), NotT),
    vts_negate([XT|VT],VTS).
vts_negate(or(VTS1,VTS2), VTSo) :- 
    vts_negate(VTS1,VTSo1),
    (
        VTSo1 == vtbot 
    -> 
        VTSo = vtbot
    ;
        vts_negate(VTS2,VTSo2),
        vtsglb(VTSo2,VTSo1,VTSo3),
        remove_subsumed_vt(VTSo3,VTSo)
    ).     


/* test0(VTS) :- 
      vts_negate([['$VAR'(1)/integer,'$VAR'(2)/list(top)],
		['$VAR'(1)/list(integer)]],VTS).
*/

/* vts_proj_forall(VTSi,NVar,VTSo) VTSo approximates from below
   \forall Vars. VTSi where Vars are set of variables that have an
   identification number greater than NVar.  */

vts_proj_forall(VTSi,NVar,VTSo) :- 
    vts_negate(VTSi,VTSne), % VTSne=~VTSi
    remove_vt_bot(VTSne,VTSneReduced), % remove redundency
    vts_restrict(VTSneReduced,NVar,VTSne1), % existential projection
    vts_negate(VTSne1,VTSo).

/*
test1(VTS) :-  vts_proj_forall([['$VAR'(1)/list(top),'$VAR'(2)/top]],1,VTS).
test2(VTS) :-  
    vts_proj_forall([['$VAR'(1)/list(top),'$VAR'(2)/list(list(top))]],1,VTS).
test3(VTS) :-  vts_proj_forall([ 
		['$VAR'(1)/list(top),'$VAR'(2)/list(top)],
		['$VAR'(1)/list(top),'$VAR'(2)/not(list(top))]],1,VTS).
*/

/* pseduo_complement(Left,Right,Complement) Complement = Left -> Right */

pseduo_complement(Left,Right,Complement) :-
    vts_negate(Left,LeftNe),          % negation
    reduce(or(Right,LeftNe),Complement).   % remove redundency

/*
test4(X) :- pseduo_complement([['$VAR'(1)/integer]], [['$VAR'(1)/integer]], X).
*/





/* rtype_solve_bw(+EQS,+VTSright,-VTSleft): reverse abstract constraint
   solving */

rtype_solve_bw(EQS,VTSRt,VTSLt) :-
    rtype_solve_bw_1(VTSRt,EQS,VTSLt). 

rtype_solve_bw_2(EQS,VT,VTS) :- 
    ( setof(VTLt,rewrite_to_solved_form(EQS,VT,VTLt),VTS) -> true; VTS=[] ).

rtype_solve_bw_1(vtbot,_EQS,vtbot). 
rtype_solve_bw_1([],_EQS,[]).
rtype_solve_bw_1([H|L],EQS,DVTS) :- 
    rtype_solve_bw_2(EQS,[H|L],VTS0),
    join_all(VTS0,DVTS1),
    remove_subsumed_vt(DVTS1,DVTS).
rtype_solve_bw_1(or(DVTS1,DVTS2),EQS,DVTS) :- 
    rtype_solve_bw_1(DVTS1,EQS,DVTSo1),
    rtype_solve_bw_1(DVTS2,EQS,DVTSo2),
    remove_vts_subsumed_by(DVTSo1,DVTSo2,DVTSo3),
    (
        DVTSo3 == vtbot 
    ->
        DVTS = DVTSo2
     ; 
        remove_vts_subsumed_by(DVTSo2,DVTSo3,DVTSo4),
        (
            DVTSo4 == vtbot
        ->
            DVTS = DVTSo3
        ;
            DVTS = or(DVTSo3,DVTSo4)
        )
    ).


/* vt_split(+Vs,+VTi,-VTo1,-VTo2) VTo1=VTi_restricted_to_Vs and
   VTo2=VTi_restricted_to_other_variables. */

vt_split([],VT,[],VT).
vt_split([V|Vs],VTi,VTo1,VTo2) :- 
    vt_split_1(VTi,[V|Vs],VTo1,VTo2).
    

vt_split_1([],_,[],[]).
vt_split_1([V/T|VT],[V|Vs],[V/T|VTo1],VTo2) :- 
    !,
    vt_split(Vs,VT,VTo1,VTo2).
vt_split_1([Eq|VT],[V|Vs],VTo1,[Eq|VTo2]) :- 
    !,
    vt_split([V|Vs],VT,VTo1,VTo2).




/* rewrite(+Eqsi,+VTi,-Eqso,-VTo) <Eqsi,VTi> rewrites
   into <Eqso,VTo> as specified in the paper */

rewrite(EQSi,VTi,EQSo,VTo) :- 
    select(Var/Tm, EQSi, EQSo), % non-deterministic
    once(vt_split([Var],VTi,_,Exists_Var_VTi)),
    vars(Tm,VarsInTm),
    once(vt_split(VarsInTm,VTi,Exists_VarsNotInTm_VTi,Exists_VarsInTm_VTi)),
    (
        (  %EQ-UP 
	       Exists_VarsNotInTm_VTi = [] 
	 -> 
	       VTo = VTi % VTi does not contrain type of Tm
	 ; 
	       type_LowApprox(Tm,VTi,TpTmLow),                
               ( 
		   TpTmLow = top 
		-> 
		   VTo = Exists_VarsInTm_VTi 
		;
	           vtglb([Var/TpTmLow],Exists_VarsInTm_VTi,VTo) 
	       )
	)
     ;
        (                                          % EQ-DOWN
	     mapsto(Var,VTi,TpVar) 
	 -> 
	     once(vts_LowApprox(TpVar,Tm,VTS)),
             select_vt(VTS,VT),                % non-deterministic
             vtglb(VT,Exists_Var_VTi,VTo)
	 ;   % TpVar = top by default
	     VTo=Exists_Var_VTi
	 )
     ;
         VTo = VTi                                     % EQ-REMOVE
    ).

select_vt(vtbot,vtbot).
select_vt([],[]). % top variable typing
select_vt([H|L],[H|L]).
select_vt(or(DVTS1,DVTS2),VT) :- 
    select_vt(DVTS1,VT) ; select_vt(DVTS2,VT).

/* rewrite_to_solved_form(+Eqsi,+VTi,-VTo) <Eqsi,VTi>
   rewrites to <[],VTo> in zero on more steps */

rewrite_to_solved_form([],VTi,VTi).
rewrite_to_solved_form([EQ|EQS],VTi,VTo) :- 
    rewrite([EQ|EQS],VTi,EQS1,VT1),
    rewrite_to_solved_form(EQS1,VT1,VTo).
        
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

tand_prim(atom,T2,T3) :- 
    tand_prim_atom(T2,T3).
tand_prim(atomic,T2,T3) :- 
    tand_prim_atomic(T2,T3).
tand_prim(integer,T2,T3) :- 
    tand_prim_integer(T2,T3).
tand_prim(float,T2,T3) :- 
    tand_prim_float(T2,T3).
tand_prim(number,T2,T3) :- 
    tand_prim_number(T2,T3).
tand_prim(not(T1),T2,T3) :- tand_prim_not_X(T1,T2,T3). 

tand_prim_not_X(atom,T2,T3) :- 
    tand_prim_not_atom(T2,T3).
tand_prim_not_X(atomic,T2,T3) :- 
    tand_prim_not_atomic(T2,T3).
tand_prim_not_X(integer,T2,T3) :- 
    tand_prim_not_integer(T2,T3).
tand_prim_not_X(float,T2,T3) :- 
    tand_prim_not_float(T2,T3).
tand_prim_not_X(number,T2,T3) :- 
    tand_prim_not_number(T2,T3).

tand_prim_atom(atom,atom) :- !.
tand_prim_atom(atomic,atom) :- !.
tand_prim_atom(integer,bot) :- !.
tand_prim_atom(float,bot) :- !.
tand_prim_atom(number,bot) :- !.
tand_prim_atom(not(atom),bot) :- !.
tand_prim_atom(not(atomic),bot) :- !.
tand_prim_atom(not(integer),atom) :- !.
tand_prim_atom(not(float),atom) :- !.
tand_prim_atom(not(number),atom) :- !.

tand_prim_atomic(atom,atom) :- !.
tand_prim_atomic(atomic,atomic) :- !.
tand_prim_atomic(integer,integer) :- !.
tand_prim_atomic(float,float) :- !.
tand_prim_atomic(number,number) :- !.
tand_prim_atomic(not(atom),number) :- !.
tand_prim_atomic(not(atomic),bot) :- !.
tand_prim_atomic(not(integer),or(atom,float)) :- !.
tand_prim_atomic(not(float),or(atom,integer)) :- !.
tand_prim_atomic(not(number),atom) :- !.


tand_prim_integer(atom,bot) :- !.
tand_prim_integer(atomic,integer) :- !.
tand_prim_integer(integer,integer) :- !.
tand_prim_integer(float,bot) :- !.
tand_prim_integer(number,integer) :- !.
tand_prim_integer(not(atom),integer) :- !.
tand_prim_integer(not(atomic),bot) :- !.
tand_prim_integer(not(integer),bot) :- !.
tand_prim_integer(not(float),integer) :- !.
tand_prim_integer(not(number),bot) :- !.


tand_prim_float(atom,bot) :- !.
tand_prim_float(atomic,float) :- !.
tand_prim_float(integer,bot) :- !.
tand_prim_float(float,float) :- !.
tand_prim_float(number,float) :- !.
tand_prim_float(not(atom),float) :- !.
tand_prim_float(not(atomic),bot) :- !.
tand_prim_float(not(integer),float) :- !.
tand_prim_float(not(float),bot) :- !.
tand_prim_float(not(number),bot) :- !.


tand_prim_number(atom,bot) :- !.
tand_prim_number(atomic,number) :- !.
tand_prim_number(integer,integer) :- !.
tand_prim_number(float,float) :- !.
tand_prim_number(number,number) :- !.
tand_prim_number(not(atom),number) :- !.
tand_prim_number(not(atomic),bot) :- !.
tand_prim_number(not(integer),float) :- !.
tand_prim_number(not(float),integer) :- !.
tand_prim_number(not(number),bot) :- !.

%

tand_prim_not_atom(atom,bot) :- !.
tand_prim_not_atom(atomic,number) :- !.
tand_prim_not_atom(integer,integer) :- !.
tand_prim_not_atom(float,float) :- !.
tand_prim_not_atom(number,number) :- !.
tand_prim_not_atom(not(atom),not(atom)) :- !.
tand_prim_not_atom(not(atomic),not(atomic)) :- !.
tand_prim_not_atom(not(integer),or(float,not(atomic))) :- !.
tand_prim_not_atom(not(float),or(integer,not(atomic))) :- !.
tand_prim_not_atom(not(number),not(atomic)) :- !.

tand_prim_not_atomic(atom,bot) :- !.
tand_prim_not_atomic(atomic,bot) :- !.
tand_prim_not_atomic(integer,bot) :- !.
tand_prim_not_atomic(float,bot) :- !.
tand_prim_not_atomic(number,bot) :- !.
tand_prim_not_atomic(not(atom),not(atomic)) :- !.
tand_prim_not_atomic(not(atomic),not(atomic)) :- !.
tand_prim_not_atomic(not(integer),not(atomic)) :- !.
tand_prim_not_atomic(not(float),not(atomic)) :- !.
tand_prim_not_atomic(not(number),not(atomic)) :- !.


tand_prim_not_integer(atom,atom) :- !.
tand_prim_not_integer(atomic,or(float,atom)) :- !.
tand_prim_not_integer(integer,bot) :- !.
tand_prim_not_integer(float,float) :- !.
tand_prim_not_integer(number,float) :- !.
tand_prim_not_integer(not(atom),or(float,not(atomic))) :- !.
tand_prim_not_integer(not(atomic),not(atomic)) :- !.
tand_prim_not_integer(not(integer),not(integer)) :- !.
tand_prim_not_integer(not(float),not(number)) :- !.
tand_prim_not_integer(not(number),not(number)) :- !.


tand_prim_not_float(atom,atom) :- !.
tand_prim_not_float(atomic,or(integer,atom)) :- !.
tand_prim_not_float(integer,integer) :- !.
tand_prim_not_float(float,bot) :- !.
tand_prim_not_float(number,integer) :- !.
tand_prim_not_float(not(atom),or(integer,not(atomic))) :- !.
tand_prim_not_float(not(atomic),not(atomic)) :- !.
tand_prim_not_float(not(integer),not(number)) :- !.
tand_prim_not_float(not(float),not(float)) :- !.
tand_prim_not_float(not(number),not(number)) :- !.


tand_prim_not_number(atom,atom) :- !.
tand_prim_not_number(atomic,atom) :- !.
tand_prim_not_number(integer,bot) :- !.
tand_prim_not_number(float,bot) :- !.
tand_prim_not_number(number,bot) :- !.
tand_prim_not_number(not(atom),not(atomic)) :- !.
tand_prim_not_number(not(atomic),not(atomic)) :- !.
tand_prim_not_number(not(integer),not(number)) :- !.
tand_prim_not_number(not(float),not(number)) :- !.
tand_prim_not_number(not(number),not(number)) :- !.

%%%%%%


tor_prim(atom,T2,T3) :- 
    tor_prim_atom(T2,T3).
tor_prim(atomic,T2,T3) :- 
    tor_prim_atomic(T2,T3).
tor_prim(integer,T2,T3) :- 
    tor_prim_integer(T2,T3).
tor_prim(float,T2,T3) :- 
    tor_prim_float(T2,T3).
tor_prim(number,T2,T3) :- 
    tor_prim_number(T2,T3).
tor_prim(not(T1),T2,T3) :-  tor_prim_not_X(T1,T2,T3).

tor_prim_not_X(atom,T2,T3) :- 
    tor_prim_not_atom(T2,T3).
tor_prim_not_X(atomic,T2,T3) :- 
    tor_prim_not_atomic(T2,T3).
tor_prim_not_X(integer,T2,T3) :- 
    tor_prim_not_integer(T2,T3).
tor_prim_not_X(float,T2,T3) :- 
    tor_prim_not_float(T2,T3).
tor_prim_not_X(number,T2,T3) :- 
    tor_prim_not_number(T2,T3).

tor_prim_atom(atom,atom) :- !.
tor_prim_atom(atomic,atomic) :- !.
tor_prim_atom(integer,or(atom,integer)) :- !.
tor_prim_atom(float,or(atom,float)) :- !.
tor_prim_atom(number,atomic) :- !.
tor_prim_atom(not(atom),top) :- !.
tor_prim_atom(not(atomic),not(number)) :- !.
tor_prim_atom(not(integer),not(integer)) :- !.
tor_prim_atom(not(float),not(float)) :- !.
tor_prim_atom(not(number),not(number)) :- !.

tor_prim_atomic(atom,atomic) :- !.
tor_prim_atomic(atomic,atomic) :- !.
tor_prim_atomic(integer,atomic) :- !.
tor_prim_atomic(float,atomic) :- !.
tor_prim_atomic(number,atomic) :- !.
tor_prim_atomic(not(atom),top) :- !.
tor_prim_atomic(not(atomic),top) :- !.
tor_prim_atomic(not(integer),top) :- !.
tor_prim_atomic(not(float),top) :- !.
tor_prim_atomic(not(number),top) :- !.


tor_prim_integer(atom,or(integer,atom)) :- !.
tor_prim_integer(atomic,atomic) :- !.
tor_prim_integer(integer,integer) :- !.
tor_prim_integer(float,number) :- !.
tor_prim_integer(number,number) :- !.
tor_prim_integer(not(atom),not(atom)) :- !.
tor_prim_integer(not(atomic),or(integer,not(atomic))) :- !.
tor_prim_integer(not(integer),top) :- !.
tor_prim_integer(not(float),not(float)) :- !.
tor_prim_integer(not(number),not(float)) :- !.


tor_prim_float(atom,or(float,atom)) :- !.
tor_prim_float(atomic,atomic) :- !.
tor_prim_float(integer,number) :- !.
tor_prim_float(float,float) :- !.
tor_prim_float(number,number) :- !.
tor_prim_float(not(atom),not(atom)) :- !.
tor_prim_float(not(atomic),or(float,not(atomic))) :- !.
tor_prim_float(not(integer),not(integer)) :- !.
tor_prim_float(not(float),top) :- !.
tor_prim_float(not(number),not(integer)) :- !.


tor_prim_number(atom,atomic) :- !.
tor_prim_number(atomic,atomic) :- !.
tor_prim_number(integer,number) :- !.
tor_prim_number(float,number) :- !.
tor_prim_number(number,number) :- !.
tor_prim_number(not(atom),not(atom)) :- !.
tor_prim_number(not(atomic),not(atom)) :- !.
tor_prim_number(not(integer),top) :- !.
tor_prim_number(not(float),top) :- !.
tor_prim_number(not(number),top) :- !.

%

tor_prim_not_atom(atom,top) :- !.
tor_prim_not_atom(atomic,top) :- !.
tor_prim_not_atom(integer,not(atom)) :- !.
tor_prim_not_atom(float,not(atom)) :- !.
tor_prim_not_atom(number,not(atom)) :- !.
tor_prim_not_atom(not(atom),not(atom)) :- !.
tor_prim_not_atom(not(atomic),not(atom)) :- !.
tor_prim_not_atom(not(integer),top) :- !.
tor_prim_not_atom(not(float),top) :- !.
tor_prim_not_atom(not(number),top) :- !.

tor_prim_not_atomic(atom,not(number)) :- !.
tor_prim_not_atomic(atomic,top) :- !.
tor_prim_not_atomic(integer,or(integer,not(atomic))) :- !.
tor_prim_not_atomic(float,or(float,not(atomic))) :- !.
tor_prim_not_atomic(number,not(atom)) :- !.
tor_prim_not_atomic(not(atom),not(atom)) :- !.
tor_prim_not_atomic(not(atomic),not(atomic)) :- !.
tor_prim_not_atomic(not(integer),not(integer)) :- !.
tor_prim_not_atomic(not(float),not(float)) :- !.
tor_prim_not_atomic(not(number),not(number)) :- !.


tor_prim_not_integer(atom,not(integer)) :- !.
tor_prim_not_integer(atomic,top) :- !.
tor_prim_not_integer(integer,top) :- !.
tor_prim_not_integer(float,not(integer)) :- !.
tor_prim_not_integer(number,top) :- !.
tor_prim_not_integer(not(atom),top) :- !.
tor_prim_not_integer(not(atomic),not(integer)) :- !.
tor_prim_not_integer(not(integer),not(integer)) :- !.
tor_prim_not_integer(not(float),top) :- !.
tor_prim_not_integer(not(number),not(integer)) :- !.


tor_prim_not_float(atom,not(float)) :- !.
tor_prim_not_float(atomic,top) :- !.
tor_prim_not_float(integer,not(float)) :- !.
tor_prim_not_float(float,top) :- !.
tor_prim_not_float(number,top) :- !.
tor_prim_not_float(not(atom),top) :- !.
tor_prim_not_float(not(atomic),not(float)) :- !.
tor_prim_not_float(not(integer),top) :- !.
tor_prim_not_float(not(float),not(float)) :- !.
tor_prim_not_float(not(number),not(float)) :- !.


tor_prim_not_number(atom,not(number)) :- !.
tor_prim_not_number(atomic,top) :- !.
tor_prim_not_number(integer,not(float)) :- !.
tor_prim_not_number(float,not(integer)) :- !.
tor_prim_not_number(number,top) :- !.
tor_prim_not_number(not(atom),top) :- !.
tor_prim_not_number(not(atomic),not(number)) :- !.
tor_prim_not_number(not(integer),not(integer)) :- !.
tor_prim_not_number(not(float),not(float)) :- !.
tor_prim_not_number(not(number),not(number)) :- !.

%%%%%%

/* type_LowApprox(+Tm,+VT,-Tp) computes -Tp such that
   \theta\in\gamma_vt(VT) if \theta(Tm)\in\gamma_t(Tp) */

type_LowApprox(V,VT,Tp) :- 
    variable(V),
    !,
    ( mapsto(V,VT,Tp1) -> type_dnf(Tp1,Tp) ; Tp=top ).

type_LowApprox(Tm,VT,Tp) :- 
    functor(Tm,Fn,Ar),
    (
        Ar =:= 0
    ->
        Tp = top % No type contstraint on Tm is induced by VT
     ;  
        Tm =.. [_|Args],
        type_LowApprox_map(Args,VT,TArgs),
        type_LowApprox_one_level(Fn/Ar,TArgs,Tp)
     ).

type_LowApprox_map([],_,[]).
type_LowApprox_map([Tm|Tms],VT,[Tp|Tps]) :- 
    type_LowApprox(Tm,VT,Tp),
    type_LowApprox_map(Tms,VT,Tps).

/* type_LowApprox_one_level(+Fn/Ar,+CTs,-CT) computes CT such that
   ti\in\gamma_t(CTs(i)) if Fn(t1,...tAr)\in\gamma_t(CT). */


type_LowApprox_one_level(Fn/Ar,RTs,RT) :- % Ar > 0
    findall(AT,type_LowApprox_one_level_one_rule(Fn/Ar,RTs,AT),ATs),
    bigtlub(ATs,RT). % if two types satisfy the spec then their union does 

type_LowApprox_one_level_one_rule(Fn/Ar,RTs,AT) :- % non-deterministic
    % Ar > 0,
    typedef(Fn/Ar,Tis,To),
    tvs_LowApprox_vec(RTs,Tis,tv_top,TVS), 
    tvs_apply(TVS,To,AT).

/* tvs_LowApprox_vec(+RTs,+Tis,+TVi,-TVo) TVo=TVi\/K such that
   RTs>=K(Tis) */

tvs_LowApprox_vec([],[],TVSi,TVSi). 
tvs_LowApprox_vec([R|Rs],[P|Ps],TVSi,TVSo) :-
    type_dnf(R,DC),
    tvs_LowApprox(DC,P,TVS1),
    tvs_glb(TVS1,TVSi,TVS2),
    tvs_LowApprox_vec(Rs,Ps,TVS2,TVSo).


/* tvs_LowApprox(+RT,+PT,-TVS) TVS(PT) <= RT specified in the paper */

tvs_LowApprox(or(C1,C2),P,TVSo) :- 
    !,
    tvs_LowApprox(C1,P,TVS1),
    (
       TVS1 == tv_top -> TVSo = tv_top 
    ;  TVS1 == tv_bot -> tvs_LowApprox(C2,P,TVSo)
    ;  tvs_LowApprox(C2,P,TVS2),
       ( 
           TVS2==tv_top -> TVSo = tv_top
       ;   TVS2==tv_bot -> TVSo = TVS1
       ;   TVSo=or(TVS1,TVS2)
       )
    ).
tvs_LowApprox(and(A,C),P,TVSo) :- 
    !,
    tvs_LowApprox(A,P,TVSa),
    tvs_LowApprox(C,P,TVSc),
    tvs_glb(TVSa,TVSc,TVSo). 
tvs_LowApprox(A,P,[P/A]) :-
    var(P), %P is a parameter in a type rule
    !.
tvs_LowApprox(not(A),P,TVSo) :- 
    !,
    negate_atom_LowApprox(A,DC),
    tvs_LowApprox(DC,P,TVSo).

tvs_LowApprox(top,P,TV) :- 
    !,
    P =..[Cons|Ps],
    functor(P,Cons,Arity),
    tops(Arity,Ones),
    make_map(Ps,Ones,TV).

tvs_LowApprox(A,P,TVS) :- 
    (
        A =..[Cons|Rs],P =..[Cons|Ps]
    ->  
        make_map(Ps,Rs,TV), TVS=TV
     ;
        TVS = tv_bot
    ).



tvs_glb(tv_bot,_,tv_bot).
tvs_glb(tv_top,DTVS,DTVS).
tvs_glb(or(DTVS11,DTVS12),DTVS2,DTVS) :- 
    tvs_glb(DTVS11,DTVS2,DTVSo1),
    (   DTVSo1==tv_top ->  DTVS=tv_top
    ;   DTVSo1==tv_bot -> tvs_glb(DTVS12,DTVS2,DTVS)
    ;   tvs_glb(DTVS12,DTVS2,DTVSo2),
        (    DTVSo2==tv_top ->  DTVS=tv_top
        ;    DTVSo2==tv_bot ->  DTVS = DTVSo1
        ;    DTVS = or(DTVSo1,DTVSo2)
        )
    ).
tvs_glb([],DTVSi,DTVSo) :- ( DTVSi==tv_bot -> DTVSo=tv_bot ; DTVSo=[] ).
tvs_glb([H|L],DTVSi, DTVSo) :- 
    tvs_glb_1(DTVSi,[H|L],DTVSo). 

tvs_glb_1(tv_bot,_,tv_bot).
tvs_glb_1(tv_top,TV,TV).
tvs_glb_1(or(DTVS1,DTVS2),TV,DTVS) :- 
    tvs_glb_1(DTVS1,TV,DTVSo1),
    (
        DTVSo1 == tv_top -> DTVS = tv_top
    ;
        DTVSo1 == tv_bot -> tvs_glb_1(DTVS2,TV,DTVS)
    ;
        tvs_glb_1(DTVS2,TV,DTVSo2),
        (
             DTVSo2 == tv_top -> DTVS = tv_top
        ;
             DTVSo2 == tv_bot -> DTVS = DTVSo1
        ;
             DTVS = or(DTVSo1,DTVSo2)
        )
    ).
tvs_glb_1([],_,[]). % [] maps every parameter to bot
tvs_glb_1([H|L],TV, TVo) :- 
    tv_glb([H|L],TV,TVo). 

/* tv_glb(+TVal1,+TVal2,-TValo) Note that the default value for a type
   parameter outside the domain of a type valuation/substitution is bot
*/

/*
tv_glb([],_,[]) :- !.
tv_glb(_,[],[]) :- !.
tv_glb([P1/T1|TVal1],[P2/T2|TVal2],TVal) :-
	(
		P1 @< P2 
	->      % skip P1
		tv_glb(TVal1,[P2/T2|TVal2],TVal)
	;
		P1 @> P2
	->      % skip P2
		tv_glb([P1/T1|TVal1],TVal2,TVal)
	;
                % narrow the type assigned to P1==P2
                tand(T1,T2,T3),
                TVal=[P1/T3|TVal12],
		tv_glb(TVal1,TVal2,TVal12)
	).
*/

tv_glb([],_,[]).
tv_glb([P1/T1|TV1],TV2,TV3) :- tv_glb_1(TV2,P1,T1,TV1,TV3).

tv_glb_1([],_,_,_,[]).
tv_glb_1([P2/T2|TV2],P1,T1,TV1,TV) :-
	(
		P1 @< P2 
	->      % skip P1
		tv_glb_1(TV1,P2,T2,TV2,TV)
	;
		P1 @> P2
	->      % skip P2
		tv_glb_1(TV2,P1,T1,TV1,TV)
	;
                % narrow the type assigned to P1==P2
                tand(T1,T2,T3),
                TV=[P1/T3|TV12],
		tv_glb(TV1,TV2,TV12)
	).

tvs_lub(tv_top,_,tv_top).
tvs_lub(tv_bot,DTVS,DTVS).
tvs_lub(or(DTVS11,DTVS12),DTVS2,DTVS) :- 
    tvs_lub(DTVS11,DTVS2,DTVSo1),
    (   DTVSo1==tv_top ->  DTVS=tv_top
    ;   DTVSo1==tv_bot -> tvs_lub(DTVS12,DTVS2,DTVS)
    ;   tvs_lub(DTVS12,DTVS2,DTVSo2),
        (    DTVSo2==tv_top ->  DTVS=tv_top
        ;    DTVSo2==tv_bot ->  DTVS = DTVSo1
        ;    DTVS = or(DTVSo1,DTVSo2)
        )
    ).
tvs_lub([],DTVSi,DTVSo) :- 
    ( DTVSi==tv_bot -> DTVSo=[] ; DTVSo=DTVSi ).
tvs_lub([H|L],DTVSi, DTVSo) :- 
    tvs_lub_1(DTVSi,[H|L],DTVSo). 

tvs_lub_1(tv_top,_,tv_top).
tvs_lub_1(tv_bot,TV,TV).
tvs_lub_1(or(DTVS1,DTVS2),TV,DTVS) :- 
    tvs_lub_1(DTVS1,TV,DTVSo1),
    (
        DTVSo1 == tv_top -> DTVS = tv_top
    ;
        DTVSo1 == tv_bot -> tvs_lub_1(DTVS2,TV,DTVS)
    ;
        tvs_lub_1(DTVS2,TV,DTVSo2),
        (
             DTVSo2 == tv_top -> DTVS = tv_top
        ;
             DTVSo2 == tv_bot -> DTVS = DTVSo1
        ;
             DTVS = or(DTVSo1,DTVSo2)
        )
    ).
tvs_lub_1([],TV,TV). % [] maps every parameter to bot
tvs_lub_1([H|L],TV, TVo) :- 
    tv_lub([H|L],TV,TVo). 

    

/* tv_lub(+TVal1,+TVal2,-TValo). Note that the default value for any
   paramter outside the domain of a type valustion is bot. */

/*
tv_lub([],TV,TV).
tv_lub([P/T|TVal],[],[P/T|TVal]).
tv_lub([P1/T1|TVal1],[P2/T2|TVal2],[P/T|TVal]) :-
	(
		P1 @< P2
	->
		P=P1,
		T=T1,
		tv_lub(TVal1,[P2/T2|TVal2],TVal)
	;
		P1 @> P2
	->
		P=P2,
		T=T2,
		tv_lub([P1/T1|TVal1],TVal2,TVal)
	;
		P=P1, 
		tor(T1,T2,T),
		tv_lub(TVal1,TVal2,TVal)
	).
*/

tv_lub([],TV,TV).
tv_lub([P1/T1|TV1],TV2,TV3) :- 
    tv_lub_1(TV2,P1,T1,TV1,TV3).

tv_lub_1([],P1,T1,TV1,[P1/T1|TV1]).
tv_lub_1([P2/T2|TV2],P1,T1,TV1,[P/T|TV]) :-
	(
		P1 @< P2
	->
		P=P1, T=T1, tv_lub_1(TV1,P2,T2,TV2,TV)
	;
		P1 @> P2
	->
		P=P2, T=T2, tv_lub_1(TV2,P1,T1,TV1,TV)
	;
		P=P1, tor(T1,T2,T), tv_lub(TV1,TV2,TV)
	).


/* tv_apply(+TSub,+TpIn,-Tp) TpO=TSub(TpIn). It unravals the input
   type TpIn in the top-down manner and constructs the output type TpO
   in the bottom-up manner. */

tv_apply(tv_top,_,top).  % special rule for tv_top
tv_apply(tv_bot,_,bot).  % special rule for tv_bot
tv_apply([],TpIn,TpO) :-
	(    
		var(TpIn) %parameter(TpIn)
	->
		TpO=bot   %default value for any parameter is bot
	; 
		TpIn =.. [F|ArgsIn],
		tv_apply_all(ArgsIn,[],ArgsO),
		TpO =.. [F|ArgsO]         
	).
tv_apply([H|TSub],TpIn,TpO) :-
	(    
		var(TpIn) %parameter(TpIn)
	->
		( 
			mapsto(TpIn,[H|TSub],TpO) 
		-> 
			true
		; 
			TpO=bot % note default value for any
				    % Tp parameter is bot
		)
	; 
		TpIn =.. [F|ArgsIn],
		tv_apply_all(ArgsIn,[H|TSub],ArgsO),
		TpO =.. [F|ArgsO]         
	).

tv_apply_all([],_,[]).
tv_apply_all([ArgIn|ArgsIn],TSub,[ArgO|ArgsO]) :-
	tv_apply(TSub,ArgIn,ArgO),
	tv_apply_all(ArgsIn,TSub,ArgsO).

type_UpApprox(V,VT,Tp) :- 
    variable(V),
    !,
    ( mapsto(V,VT,Tp1) -> type_dnf(Tp1,Tp) ; Tp=top ).

type_UpApprox(Tm,VT,Tp) :- 
    functor(Tm,Fn,Ar),
    Tm =.. [_|Args],
    type_UpApprox_map(Args,VT,TArgs),
    type_UpApprox_one_level(Fn/Ar,TArgs,Tp).


type_UpApprox_map([],_,[]).
type_UpApprox_map([Tm|Tms],VT,[Tp|Tps]) :- 
    type_UpApprox(Tm,VT,Tp),
    type_UpApprox_map(Tms,VT,Tps).

/*
type_UpApprox_one_level(Fn/Ar,CTs,CT) :- 
    findall(AT,type_UpApprox_one_level_one_rule(Fn/Ar,CTs,AT),ATs),
    bigtglb(ATs,CT).

type_UpApprox_one_level_one_rule(Fn/0,[],AT)   :- % non-deterministic 
%    prim_type(AT),  % extra clause for constant and types 
%    membership_prim(AT,Fn). 
%    Smallest builtin type needs be returned since types will be conjuncted
    (  integer(Fn) -> AT = integer
     ; float(Fn)   -> AT = float
     ; atom(Fn)    -> AT = atom 
    ).

type_UpApprox_one_level_one_rule(Fn/Ar,CTs,AT) :- % non-deterministic 
    typedef(Fn/Ar,Tis,To),
    tvs_UpApprox_vec(CTs,Tis,[],TVS), % [] maps every parameter to bot
    tvs_apply(TVS,To,AT).
*/


type_UpApprox_one_level(Fn/Ar,CTs,CT) :- 
    ( 
         bagof(AT,type_UpApprox_one_level_one_rule(Fn/Ar,CTs,AT),ATs)
     -> 
         bigtglb(ATs,CT) % At least one type rule applies to Fn/Ar and CTs
     ;  integer(Fn) -> CT = integer
     ;  float(Fn)   -> CT = float
     ;  atom(Fn)    -> CT = atom  
     ;  CT = top % no applicalble rules, not belonging to a primitive type
    ).

%type_UpApprox_one_level_one_rule(Fn/0,[],AT)   :- % non-deterministic 
%    prim_type(AT),  % extra clause for constant and types 
%    membership_prim(AT,Fn). 
%    Smallest builtin type needs be returned since types will be conjuncted
%    (  integer(Fn) -> AT = integer
%     ; float(Fn)   -> AT = float
%     ; atom(Fn)    -> AT = atom 
%    ).

type_UpApprox_one_level_one_rule(Fn/Ar,CTs,AT) :- % non-deterministic 
    typedef(Fn/Ar,Tis,To),
    tvs_UpApprox_vec(CTs,Tis,[],TVS), % [] maps every parameter to bot
    tvs_apply(TVS,To,AT).

/* tvs_UpApprox_vec(+RTs,+Tis,+TVSi,-TVSo) TVSo = TVi \/ K such that
   RTs <= K(Tis) */

tvs_UpApprox_vec([],[],TVSi,TVSi). 
tvs_UpApprox_vec([R|Rs],[P|Ps],TVSi,TVSo) :-
    type_dnf(R,DC), % make sure tvs_UpApprox is called with first arg in dnf
    tvs_UpApprox(DC,P,TVS1),
    tvs_lub(TVS1,TVSi,TVS2),
    tvs_UpApprox_vec(Rs,Ps,TVS2,TVSo).


/* tvs_UpApprox(+RT,+Poly,-TVS) TVS(Poly)>=RT specified in the paper*/


tvs_UpApprox(or(A,C),P,TVSo) :- 
    !,
    tvs_UpApprox(A,P,TVS1),
    (
       TVS1==tv_top -> TVSo = tv_top 
    ;  TVS1==tv_bot -> tvs_UpApprox(C,P,TVSo)
    ;  tvs_UpApprox(C,P,TVS2),
       ( 
           TVS2==tv_top -> TVSo = tv_top
       ;   TVS2==tv_bot -> TVSo = TVS1
       ;   TVSo=or(TVS1,TVS2)
       )
    ).
tvs_UpApprox(and(A,C),P,TVSo) :- 
    !,
    tvs_UpApprox(A,P,TVSa),
    tvs_UpApprox(C,P,TVSc),
    tvs_glb(TVSa,TVSc,TVSo). 
tvs_UpApprox(A,P,[P/A]) :-
    var(P), %P is a parameter in a type rule
    !.
tvs_UpApprox(not(A),P,TVSo) :- 
    !,
    negate_atom_UpApprox(A,DC),
    tvs_UpApprox(DC,P,TVSo).
tvs_UpApprox(top,_,tv_top) :-!.
tvs_UpApprox(bot,_,tv_bot) :-!.

tvs_UpApprox(A,P,TVS) :- 
    (
        A =..[Cons|Rs],P =..[Cons|Ps]
    ->  
        make_map(Ps,Rs,TVal), TVS=TVal
     ;
        TVS = tv_top
    ).


tvs_apply(tv_top,_,top).
tvs_apply(tv_bot,_,bot).
tvs_apply([],PT,RT) :- 
    tv_apply([],PT,RT).
tvs_apply([H|L],PT,RT) :- 
    tv_apply([H|L],PT,RT). 
tvs_apply(or(TVS1,TVS2),PT,RT) :- 
    tvs_apply(TVS1,PT,RT1),
    (    RT1 == top -> RT = top
    ;    RT1 == bot -> tvs_apply(TVS2,PT,RT)
    ;    tvs_apply(TVS2,PT,RT2),
         (   RT2 == top -> RT = top
         ;   RT2 == bot -> RT = RT1
         ;   tor(RT1,RT2,RT)
         )
    ).

/*

test_solve :- numbervars((B,C,D,E,F,G,H,I,J,K,L,M,N),1,_),
	rtype_solve([J/I,K/G,L/C,M/D,N/E],[B/list(atom),G/atom,H/list(atom),I/list(atom),J/list(atom),K/atom,M/atom,N/list(atom)]or[B/list(atom),G/atom,H/list(atom),I/list(atom),J/list(atom),K/list(atom),M/list(atom),N/list(atom)],_).

*/
